# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini@gmail.com
# ---------------------------------------------------------------------------------------------------------------

# ---------------------------------------------------------------------------------------------------------------
# Management of exceptions for piwiPre
# ---------------------------------------------------------------------------------------------------------------

# raise PwpError(): prints a message and generate an exception


import inspect
import time
import datetime
import sys


class PwpLog:
    def __init__(self):
        logfile_name = time.strftime("piwiPre_%Y_%m_%d.log")
        self.logfile = open(logfile_name, "a", encoding="utf-8")
        self.start_time = datetime.datetime.now()
        self.print_debug = False
        self.msg(f"---- piwiPre start {self.start_time}")
        self.pictures_nb = 0
        self.data = {           # normal logs are not stored
            'info': [],         # logs that are stored for tests
            'Warning': [],      # a problem that do not require stopping the program
            'ERROR': [],        # under normal circumstances, stops the program. Can be trapped for test
        }

    def configure(self, config):
        self.print_debug = config['debug']

    def reset_data(self):
        self.start_time = datetime.datetime.now()
        self.data = {
            'info': [],
            'Warning': [],
            'ERROR': [],
        }
        nb = self.pictures_nb
        self.pictures_nb = 0
        return nb

    def msg_nb(self, level):
        return len(self.data[level])

    def end(self):
        end = datetime.datetime.now()
        self.msg(f"--- end          = {end} ---")
        self.msg(f"--- Nb files     = {self.pictures_nb}")
        self.msg(f"--- duration     = {end - self.start_time}")
        if self.pictures_nb:
            self.msg(f"--- duration/pic = {(end - self.start_time) / self.pictures_nb}")
        self.msg("------------------------------------")

    def incr_picture(self):
        self.pictures_nb += 1

    def do_msg(self, msg, context=None, level='msg', flush=False):
        if context is not None:
            if level not in ['debug', 'info'] or self.print_debug:
                print(f"{level:7} {context}")
                print(f"{level:7} {context}", file=self.logfile)

        if level not in ['debug', 'info'] or self.print_debug:
            print(f"{level:7} {msg}")
            print(f"{level:7} {msg}", file=self.logfile)
            if flush:
                sys.stdout.flush()

        if level != 'msg' and level != 'debug':
            self.data[level].append(msg)

        if level == 'warning' and self.msg_nb('warning') > 20:
            raise PwpError("Too much warnings")

        if level == 'ERROR' and self.msg_nb('ERROR') > 20:
            raise PwpFatal("Too much errors, aborting")

    def warning(self, msg, context=None):        # warning is always kept for test, and ALWAYS printed
        self.do_msg(msg, context=context, level='Warning', flush=True)

    def msg(self, msg, context=None):           # info is NOT kept for test, and ALWAYS printed
        self.do_msg(msg, context=context, level='msg')

    def info(self, msg, context=None):          # info is always kept for test, and printed only if --debug
        self.do_msg(msg, context=context, level='info', flush=True)

    def debug(self, msg, context=None):         # debug is NOT kept for test, and printed only if --debug
        self.do_msg(msg, context=context, level='debug', flush=True)

    # error must NOT be declared, it is mandatory to go through PwpError
    # def error(self, msg, context=None):
    #     self.do_msg(msg, context=context, level='ERROR', flush=True)


class PwpError(Exception):
    def __init__(self, msg, context=None):
        LOGGER.do_msg(msg, level='ERROR', context=context, flush=True)


class PwpInternalError(PwpError):
    def __init__(self, msg: str):
        previous_frame = inspect.currentframe().f_back
        (filename, line_number, _function_name, _lines, _index) = inspect.getframeinfo(previous_frame)
        context = f"{filename}:{line_number:3}"
        super().__init__("INTERNAL " + msg, context=context)


class PwpConfigError(PwpError):
    def __init__(self, msg, context="Cmd-line/configuration "):
        super().__init__(msg, context=context)


# class PwpExit(PwpError):
#     def __init__(self, msg):
#         super().__init__(msg, context="Exiting from main")


class PwpFatal(Exception):  # should not be trapped
    def __init__(self, msg):
        LOGGER.do_msg(msg, level='FATAL', flush=True)


LOGGER = PwpLog()



# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini@gmail.com
# ---------------------------------------------------------------------------------------------------------------

import datetime
import logging
import re
import os

# pip install Pillow
from PIL import Image
from PIL.ExifTags import TAGS
from PIL import ImageOps

# pip install iptcinfo3
from iptcinfo3 import IPTCInfo

from piwiPre.pwpActor import ACTOR
from piwiPre.pwpConfig import PwpConfig
from piwiPre.pwpErrors import LOGGER, PwpConfigError


class PwpObject:
    """Root class for PwpJpg and PwpMP4"""

    def __init__(self, filename):
        self.filename = filename
        self.copyright = None
        self.special = None
        self.author = None

    def close(self):
        pass  # nothing to do on txt files

    def write_metadata(self, _new_author, _new_copyright, _new_special):
        pass  # nothing to do on txt files

    def rotate(self, _stage: str, _any):
        return False  # nothing to do with text or mp4

    @staticmethod
    def guess_file_date(image_path: str, config: PwpConfig):
        """
        :param image_path: the full path to image
        :param config: the current configuration
        :return: date
        """

        dates = config['dates']
        if not isinstance(dates, dict):
            dates = {}
        if dates is not None and 'NO-DATE' in dates:
            spec = dates['NO-DATE']
            if 'forced' not in spec:
                raise PwpConfigError("date:NO-DATE config without a 'forced' statement")
            new_date = spec['forced']
            return datetime.datetime(**new_date)

        filename = os.path.basename(image_path)
        template = os.path.basename(config['names'])
        file_info = ACTOR.get_info_from_format(template, filename)
        if file_info is not None and 'Y' in file_info and file_info['Y'] is not None and \
                'm' in file_info and file_info['m'] is not None and \
                'd' in file_info and file_info['d'] is not None and \
                'H' in file_info and file_info['H'] is not None and \
                'S' in file_info and file_info['S'] is not None:
            return datetime.datetime(year=int(file_info['Y']),
                                     month=int(file_info['m']),
                                     day=int(file_info['d']),
                                     hour=int(file_info['H']),
                                     minute=int(file_info['M']),
                                     second=int(file_info['S']))  # TODO: Manage timezone and AM/PM

        m = re.match(r"VID(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)\.mp4", filename) or \
            re.match(r"VID_(\d\d\d\d)(\d\d)(\d\d)_(\d\d)(\d\d)(\d\d)\.mp4", filename) or \
            re.match(r".*-(\d\d\d\d)-(\d\d)-(\d\d)-(\d\d)h(\d\d)-(\d\d).*\.jpg", filename)
        if m:
            year = int(m.group(1))
            month = int(m.group(2))
            day = int(m.group(3))
            hours = int(m.group(4))
            minutes = int(m.group(5))
            seconds = int(m.group(6))
            return datetime.datetime(year=year, month=month, day=day, hour=hours, minute=minutes, second=seconds)

        m = re.match(r"IMG-(\d\d\d\d)-(\d\d)(\d\d)-W.(\d\d)(\d\d).*\.jpg", filename) or \
            re.match(r"IMG-(\d\d\d\d)(\d\d)(\d\d)-W.(\d\d)(\d\d).*\.jpg", filename)
        # for WhatsApp images, which does not have IPTC date information, but get a date in the picture name
        if m:
            year = int(m.group(1))
            month = int(m.group(2))
            day = int(m.group(3))
            hours = int(m.group(4)) + 12
            minutes = int(m.group(5))
            seconds = 0
            return datetime.datetime(year=year, month=month, day=day, hour=hours, minute=minutes, second=seconds)

        LOGGER.debug(f"Cannot guess date for {image_path}")
        return datetime.datetime.fromtimestamp(os.path.getmtime(image_path))

    def set_metadata(self, _stage, config: PwpConfig,
                     date=datetime.datetime.today(),
                     new_author="Photographer",
                     confirm=True):
        """
        :param _stage: triage or album
        :param config: local configuration
        :param date: date to set
        :param new_author: author to set in exif
        :param confirm: is True, always write metadata, if False, write only if not set
        :return: True if the file was modified
        """
        self.close()  # we know that image will be no more modified

        if config['enable-metadata'] is False:
            return False

        info_format = config.format('instructions')
        info_dico = config.format_dict(date, new_author)
        new_special = info_format.format(**info_dico)
        if confirm:
            modified = self.special is None or self.special == ''
        else:
            modified = self.special != new_special

        if confirm:
            modified = modified or (self.author is None or self.author == '')
        else:
            modified = modified or self.author != new_author

        info_format = config.format('copyright')
        info_dico = config.format_dict(date, new_author)
        new_copyright = info_format.format(**info_dico)
        if confirm:
            modified = modified or (self.copyright is None or self.copyright == '')
        else:
            modified = modified or self.copyright != new_copyright

        if not modified:
            return False

        # LOGGER.msg(f"write IPTC info '{self.filename}'")
        if config['dryrun']:
            return modified

        return self.write_metadata(new_author, new_copyright, new_special)


class PwpJpg(PwpObject):
    logger = logging.getLogger('iptcinfo')
    logger.setLevel(logging.ERROR)
    tags_reverse = dict(((v, k) for k, v in TAGS.items()))
    orientations = ((None, False),  # 0: is an error
                    (None, False),  # 1: nothing to do
                    (Image.FLIP_LEFT_RIGHT, False),  # 2
                    (Image.ROTATE_180, False),  # 3
                    (Image.FLIP_TOP_BOTTOM, False),  # 4
                    (Image.ROTATE_270, True),  # 5
                    (Image.ROTATE_270, False),  # 6
                    (Image.ROTATE_90, True),  # 7
                    (Image.ROTATE_90, False))  # 8
    orient_names = (('None', 'No flip'),  # 0: is an error
                    ('None', 'No flip'),  # 1: nothing to do
                    ('FLIP_LEFT_RIGHT', 'No flip'),  # 2
                    ('ROTATE_180', 'No flip'),  # 3
                    ('FLIP_TOP_BOTTOM', 'No flip'),  # 4
                    ('ROTATE_270', 'Flip'),  # 5
                    ('ROTATE_270', 'No flip'),  # 6
                    ('ROTATE_90', 'Flip'),  # 7
                    ('ROTATE_90', 'No flip'))  # 8

    def __init__(self, filename=None, image=None, iptc_info=None):
        super().__init__(filename)
        self.image = image
        self.exif_data = None
        self.DateTimeOriginal = None
        self.DateTime = None
        self.Make = None
        self.Model = None
        self.SubSecTime = None
        self.orientation = 1
        self.iptc_info = None

        if filename is not None:
            self.image = Image.open(filename)

        if self.image is None:
            LOGGER.warning(f"error opening jpg image {filename}")

        # the values initialized here are the INITIAL values BEFORE piwiPre actions
        self.exif_data = self.image.getexif()
        if self.exif_data:
            self.DateTimeOriginal = self.get_exif_value('DateTimeOriginal')
            self.DateTime = self.get_exif_value('DateTime')
            self.Make = self.get_exif_value('Make')
            self.Model = self.get_exif_value('Model')
            self.SubSecTime = self.get_exif_value('SubsecTime')  # noqa
            orient = self.get_exif_value('Orientation')
            self.orientation = int(orient) if orient is not None else 1

            # sst2 = self.get_data(exif_data, 'SubsecTimeOriginal')     # noqa
            # sst3 = self.get_data(exif_data, 'SubsecTimeDigitized')    # noqa
        else:
            LOGGER.debug(f"No exif data in '{self.filename}'")

        if filename:
            self.iptc_info = IPTCInfo(self.filename, force=True)
        else:
            self.iptc_info = iptc_info

        if isinstance(self.iptc_info, IPTCInfo):
            b1 = self.iptc_info['special instructions']
            self.special = b1.decode('utf8') if b1 is not None else None
            b2 = self.iptc_info['copyright notice']
            self.copyright = b2.decode('utf8') if b2 is not None else None
            b3 = self.iptc_info['by-line']
            self.author = b3.decode('utf8') if b3 is not None else None



        # self.close() # NO: we keep the file open because we may want to rotate it.

    def reopen(self):
        if self.image is None:
            self.image = Image.open(self.filename)

    def close(self):
        """close the image WITHOUT saving"""
        if self.image:
            self.image.close()
        self.image = None

    def save(self, filename, config: PwpConfig):
        self.filename = filename
        if not config['dryrun']:
            self.image.save(filename, exif=self.exif_data)
        self.image.close()
        if os.path.isfile(filename + '~'):
            os.remove(filename + '~')

    def get_exif_value(self, tag_id):
        # https://www.thepythoncode.com/code/extracting-image-metadata-in-python
        # get the tag name, instead of human unreadable tag id
        tag = self.tags_reverse.get(tag_id)
        data = self.exif_data.get(tag)
        # decode bytes
        if isinstance(data, bytes):
            data = data.decode()
        return data

    def set_exif_data(self, tag_id, value):
        # https://www.thepythoncode.com/code/extracting-image-metadata-in-python
        # get the tag name, instead of human unreadable tag id
        tag = self.tags_reverse.get(tag_id)
        self.exif_data[tag] = value

    def get_jpg_new_date(self, config: PwpConfig):
        """
        :param config: current configuration
        :return: date, author
        """

        author = config.default_author()

        if self.exif_data is None:
            return self.guess_file_date(self.filename, config), author

        stamp = self.DateTimeOriginal or self.DateTime
        if not stamp:
            LOGGER.debug(f"No exif dates in '{self.filename}'")
            return self.guess_file_date(self.filename, config), author

        m = re.match(r"(\d\d\d\d):(\d\d):(\d\d) (\d\d):(\d\d):(\d\d)", stamp)
        if not m:
            LOGGER.debug(f"ERROR reading date from filename '{self.filename}'")
            return self.guess_file_date(self.filename, config), author

        (year, month, day, h, m, s) = m.groups()  # noqa
        year = int(year)
        month = int(month)
        day = int(day)
        h = int(h)
        m = int(m)
        s = int(s)

        apn = self.get_exif_value('Model')
        if apn is not None:
            apn = re.sub(r"\t", "", apn)
            apn = re.sub(r"\s", "", apn)
            apn = re.sub(r"\x00", "", apn)

        LOGGER.debug(f"Date = {year:04}/{month:02}/{day:02} {h:02}:{m:02}:{s:02} before correction")
        photo_date = datetime.datetime(year=year, month=month, day=day, hour=h, minute=m, second=s)

        author = config.author(apn, photo_date)
        new_date = config.fix_date(self.filename, photo_date, apn)

        LOGGER.debug(f"Date = {new_date.year:04}/{new_date.month:02}/{new_date.day:02} " +
                     f"{new_date.hour:02}:{new_date.minute:02}:{new_date.second:02} after correction")

        return new_date, author

    def write_metadata(self, new_author, new_copyright, new_special):
        if not self.iptc_info:
            LOGGER.debug(f"ERROR writing IPTC from '{self.filename}'")
            return False

        self.iptc_info['by-line'] = new_author
        self.iptc_info['special instructions'] = new_special
        self.iptc_info['copyright notice'] = new_copyright

        self.iptc_info.save()
        ACTOR.delete(self.filename + '~')
        return True

    @staticmethod
    def get_remote_filename(config, filename):
        """
        :param config: local configuration
        :param filename: local name
        :return: the remote path to be used for sftp
        """
        win_rel_path = os.path.relpath(filename, config['web'])
        rel_path = win_rel_path.replace('\\', '/')
        remote_file = config['remote-web'] + '/' + rel_path
        return remote_file

    def thumbnail(self, filename, width, height, crop, config: PwpConfig, allow_clobber: bool):
        """
        Build a thumbnail
        :param filename: filename name of the thumbnail
        :param width: expected max width
        :param height: expected max height
        :param crop: boolean
        :param config: current configuration
        :param allow_clobber: True, False
        :return: True if the thumbnail has been generated
        """

        if os.path.isfile(filename):
            if allow_clobber:
                ACTOR.delete(filename)
            else:
                return False

        remote = (config['enable-remote-copy'] and config['remote-web'] and
                  config['remote-host'] and config['remote-port'])
        remote_file = None

        if remote:
            remote_file = self.get_remote_filename(config, filename)
            if ACTOR.remote_file_or_dir_exists(remote_file) and not allow_clobber:
                LOGGER.info(f"Thumbnail {filename} already exists remotely as {remote_file}")
                return False

        LOGGER.info(f"Thumbnail {width}x{height} crop={crop} {filename}")

        if not crop:
            im_resized = PwpJpg(image=self.image.copy(), iptc_info=self.iptc_info)
            im_resized.image.thumbnail((width, height), resample=Image.LANCZOS)
        else:
            im_resized = PwpJpg(image=ImageOps.fit(self.image, (width, height)), iptc_info=self.iptc_info)

        dir_name = os.path.dirname(filename)

        ACTOR.mkdirs(dir_name)
        im_resized.set_exif_data('ImageWidth', im_resized.image.width)
        im_resized.set_exif_data('ImageLength', im_resized.image.height)
        im_resized.save(filename, config)

        info_resized = IPTCInfo(filename, force=True)
        info_resized['copyright notice'] = self.iptc_info['copyright notice']
        info_resized['special instructions'] = self.iptc_info['special instructions']
        info_resized.save()
        ACTOR.delete(filename + '~')
        ACTOR.delete(self.filename + '~')

        if remote_file:
            remote_path = os.path.dirname(remote_file)
            LOGGER.info(f"Thumbnail {filename} sftp to {remote_path}")
            ACTOR.remote_put(filename, remote_path)

        return True

    def rotate(self, stage: str, config: PwpConfig):
        """ return True if picture was modified"""
        if config['enable-rotation'] is False:
            return False

        modified = False
        (angle, flip) = self.orientations[self.orientation]
        (a, r) = self.orient_names[self.orientation]

        new_image = self
        if angle is not None:
            if config['dryrun']:
                LOGGER.debug(f"Would clean ROTATION {self.orientation} {a} {r} '{self.filename}'")
            else:
                self.reopen()
                transposed = self.image.transpose(method=angle)
                new_image = PwpJpg(image=transposed, iptc_info=self.iptc_info)
                new_image.set_exif_data('ImageWidth', self.image.width)
                new_image.set_exif_data('ImageLength', self.image.height)
                new_image.set_exif_data('Orientation', 1)
                modified = True
                LOGGER.debug(f"ROTATION {self.orientation} {a} {r} '{self.filename}'")

        if flip:
            if config['dryrun']:
                LOGGER.debug(f"Would clean FLIP {self.orientation} {a} {r}'{self.filename}'")
            else:
                new_image.reopen()
                new_image = PwpJpg(image=new_image.image.transpose(method=Image.FLIP_LEFT_RIGHT),
                                   iptc_info=self.iptc_info)
                new_image.set_exif_data('Orientation', 1)
                modified = True
                LOGGER.debug(f"FLIP {self.orientation} {a} {r}'{self.filename}'")

        if modified and not config['dryrun']:
            new_image.save(self.filename, config)
        return modified



# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini(at)gmail.com
# ---------------------------------------------------------------------------------------------------------------

import sys
import os.path
import pathlib
import re
import pprint
import tempfile
import datetime

from piwiPre.pwpActor import ACTOR
from piwiPre.pwpParser import PwpParser
from piwiPre.pwpConfig import PwpConfig
from piwiPre.pwpJpg import PwpJpg, PwpObject
from piwiPre.pwpMp4 import PwpMp4
from piwiPre.pwpErrors import PwpError, PwpConfigError, LOGGER
from piwiPre.pwpVersion import PwpVersion
from piwiPre.pwpLicence import PwpLicence

# -----------------------------------------------------------------------------------
# Requirements, General behavior
#
# REQ 0001: piwiPre is configured with HOME/.piwiPre.ini and piwiPre.ini files found in the hierarchy of directories
# REQ 0002: piwiPre is also configured by cmdline arguments
# REQ 0003: piwiPre renames .jpg, .mp4, .txt files found in 'triage' (enable-rename) (program_24)
# REQ 0004: piwiPre inserts metadata in .jpg and .mp4 files (enable-metadata)
# REQ 0005: piwiPre generates piwigo metadata (enable-thumbnails)
# REQ 0006: piwiPre synchronizes piwigo albums that were modified (enable-piwigo-sync)
# REQ 0007: piwiPre runs on windows and Linux development stations
# REQ 0008:  --dump-config folder allows to debug the configuration hierarchy (program_2)

# REQ 0009: piwiPre configures automatically album by ADDING piwiPre.ini files (auto-configuration) (program_6, 7, 8, 9)
# REQ 0010: hand-writing a piwiPre.ini in albums may prevent any modification of files
# REQ 0011: during renaming, the name of .txt files is not changed. They simply go to the appropriate folder
# REQ 0012: tox is run on gitlab, generates html doc and coverage

# REQ 0020: 'enable-XXX' : false: feature is never active
# REQ 0021: 'enable-XXX' : true: feature is done in TRIAGE, and in ALBUMS if the result is not present


# REQ 0049: requirements are parsed from the source and put automatically in doc

# REQ 0050: temporary files are put in a real temporary dir
# REQ 0051: parser: remove trailing / at end of directories, this is common error

# REQ 0100: Renaming is based on the 'names' configuration template
# REQ 0101: Renaming takes into account the date/time of the picture shooting, found in metadata
# REQ 0102: if renaming a file clobber an existing file, piwiPre increments the last numerical field to avoid conflict


# REQ 0200: piwiPre verifies the albums are aligned with configuration (verify-albums)
# REQ 0201: piwiPre realigns pictures rotation
# REQ 0202: piwiPre updates metadata
# REQ 0203: piwiPre generates lacking thumbnails from album
# REQ 0204: album modified pictures are saved in 'BACKUP'
# REQ 0205: piwiPre removes useless thumbnails
# REQ 0206: the albums to verify are specified by 'verify-albums'
# REQ 0207: insert the 'author' metadata
# REQ 0208: use XMP metadata (piwiPre version 2.0)

# REQ 0300: generate an error if HOME/.piwiPre.ini is not protected: CANCELLED, because chmod is limited in windows
# TODO REQ 0301: piwiPre reorders the piwigo directories with the right order !  (future piwiPre version 2.0)

# -----------------------------------------------------------------------------------
# Requirements, Testing

# REQ 3000: autotest available with tox and pytest
# REQ 3001: piwipre> tests\\pwp_test.py -n number # runs the test number n                          # noqa
# REQ 3002: pwp_test -n 0 # runs all the tests
# REQ 3003: pwp_test can run all the autotests
# REQ 3004: PyCharm tests cases are saved in project files
# REQ 3005: all automatic tests run by pwp_test can be run by anyone on any development server
# REQ 3006: pwp_test assumes a valid configuration in HOME/.piwiPre.ini
# REQ 3007: coverage is run automatically by tox
# REQ 3008: coverage > 92 %
# REQ 3009: tests --dryrun, with triage and album
# REQ 3010: test error generation

# REQ 3100: test jpg files in TRIAGE, with/without rotation and metadata
# REQ 3101: test .txt files in TRIAGE
# REQ 3102: test .mp4 files in TRIAGE with metadata
# REQ 3104: tests end-2-end with piwogo sync (program_13)
# REQ 3105: test --version
# REQ 3106: test --chdir
# REQ 3107: test --licence
# REQ 3108: autotest of PwpPatcher
# REQ 3109: autotest of PwpPatcher should cover 80% of code
# REQ 3110: autotest of PwpMP4
# REQ 3111: There are no explicit remote locations on server, use config instead
# REQ 3112: someone with a valid config would run the same test on a different configuration
# REQ 3113: test --dump-config folder
# REQ 3114: test PwpError generation (program_26)
# REQ 3115: on argument error, raise an exception , not trapped
# REQ 3116: test the insertion of the 'author' metadata


# REQ 3120: test --enable-rename false  (program_27)
# REQ 3122: test dates (all cases) :  program_24
# TODO REQ 3123: test unsupported (e.g. .foo) files in TRIAGE

# TODO REQ 3200: test manual piwiPre.ini in album that prevents from renaming
# TODO REQ 3201: test manual piwiPre.ini in album that prevents from changing metadata
# TODO REQ 3202: test manual piwiPre.ini in album that prevents from rotation


# DONE: test --test-piwigo-synchronization

# DONE: a modified copy of a picture (so same metadata!) is renamed thanks to collision avoiding

# DOC: piwigo user and password in HOME .piwiPre.ini
# Doc:  HOME/.piwiPre.ini contains only server-side information
# Doc: warning: .ini files should be written in UTF8 encoding !
# Doc: usage: photo is accessible or synchronized, not web: this is not an issue, piwigo will generate thumbnails

# REQ 4000: manage remote WEB, accessible with ssh
# REQ 4001: setup ssh/sftp session
# REQ 4002: We assume that thumbnails are always coming from the corresponding file in ALBUM
# REQ 4003: a thumbnail is created only if the file in ALBUM is new, or the thumbnail was non-existent
# REQ 4004: thumbnails are created in WEB, then copied to remote-web

# ----------------------------- FUTURE ROADMAP -------------------------------------
# Future V2.0
#
#
# TODO REQ 9000: piwiPre runs on a synology NAS
# TODO REQ 9001: piwiPre runs on a workstation connected to the piwigo server through ssh, without shared directories
# TODO REQ 9002: piwiPre uses MD5 checksums to compare local and remote files
# TODO REQ 9100: piwiPre generates synology @eaDir thumbnails to speedup processing
# TODO: Parse requirements: reorder by REQ number, yell if duplicate number. non REQ are listed in their arrival order
# TODO: manage several toplevel albums inside piwigo


class PwpMain:
    allowed_chars = r"a-zA-Z0-9\-_.&@~!,;+°()àâäéèêëïîôöùûüÿçñÀÂÄÉÈÊËÏÎÔÖÙÛÜŸÇÑ "    # noqa

    def __init__(self, arguments=None):
        self.initial_cwd = os.getcwd()
        self.parser = PwpParser(arguments=arguments, program="piwiPre", with_config=True)
        self.parser_config = self.parser.config
        self.dumped_config = None      # The config reached by --dump-config, used for test only
        self.dirs_to_synchronize = []
        self.expected_thumbnails = []  # list of expected thumbnails, when processing ALBUM. reset for each dir.
        # self.cached_dirs = {}

    @staticmethod
    def get_base_from_dir(dir_name, p_config):
        base = re.sub(r".jpg$", "", dir_name)
        base = re.sub(r'[^' + PwpMain.allowed_chars + ']', "_", base, re.IGNORECASE)

        # Maybe we are processing a directory which is already the result of a renaming process
        # so its format should be the same then the directory part of 'names'

        template = os.path.basename(os.path.dirname(p_config['names']))
        res = ACTOR.get_info_from_format(template, dir_name)
        if res and res['base'] is not None and res['base'] != '':
            return res['base']
        return base

    @staticmethod
    def build_path(base, common_path: str, path):
        if base is None:
            return None
        result = base + '/' + common_path + ('' if common_path == '' else '/') + path
        if result[-1] == '/':
            result = result[:-1]
        return result

    # @staticmethod
    # def increment_last_number(filename):
    #     m = re.match(r"(.*)(\d+)(\D*)\.jpg", filename)
    #     if m is not None:
    #         prefix = m.group(1)
    #         number = int(m.group(2))
    #         postfix = m.group(3)
    #         return prefix + str(number+1) + postfix + '.jpg'
    #     else:
    #         LOGGER.msg(f"file {filename} has no integer to increment")
    #         return filename[:-4] + '01.jpg'

    def build_all_thumbs(self, target: str, config: PwpConfig, common_path: str, stage: str,
                         allow_clobber: bool):
        # returns the description of work done
        if config['enable-thumbnails'] is False:
            LOGGER.debug(f"{stage} {target} enable-thumbnails is False")
            return ''

        f = os.path.basename(target)[:-4]  # skip .jpg

        thumbs_base = config['web'] + '/' + common_path

        im = PwpJpg(target)
        thumbs = config['piwigo-thumbnails']

        summary = ''
        for name, values in thumbs.items():
            width = values['width']
            height = values['height']
            crop = values['crop']
            thumb_name = name.format(f=f)
            local_name = thumbs_base + '/' + thumb_name
            self.expected_thumbnails.append(local_name)

            if config['dryrun']:
                # CAVEAT: conflict management has NOT occurred, because the files are not created in ALBUM
                #         So we do *not* know te exact thumbnail name
                #         hence the message is not accurate
                LOGGER.info(f"Would create Thumbnail {width}x{height} crop={crop} for {target}")
            else:
                im.thumbnail(local_name, width, height, crop, config, allow_clobber)
                if summary == '':
                    summary = ' Thumbs('
                summary += name[4:-4] + ', '
        if summary != '':
            summary = summary[:-2] + ') '

        index_base = 'index.htm'
        local_name = thumbs_base + '/' + index_base
        if local_name not in self.expected_thumbnails:
            self.expected_thumbnails.append(local_name)

        if not os.path.isfile(local_name):
            if config['dryrun']:
                LOGGER.info(f"would create thumbnail html index '{local_name}'")
            else:
                LOGGER.info(f"Create thumbnail html index '{local_name}'")
                with open(local_name, 'w', encoding="utf-8") as f:
                    print("Not allowed!", file=f)
                summary += 'Index '
                remote = (config['enable-remote-copy'] and config['remote-web'] and
                          config['remote-host'] and config['remote-port'])
                if remote:
                    remote_file = PwpJpg.get_remote_filename(config, local_name)
                    remote_path = os.path.dirname(remote_file)
                    LOGGER.info(f"Index file {local_name} sftp to {remote_path}")
                    ACTOR.remote_put(local_name, remote_path)
        return summary

    @staticmethod
    def get_cached_folder(path) -> [str]:
        # if path in self.cached_dirs:
        #     # update_with_dir has already been done, no need to update again
        #     return self.cached_dirs[path]
        album_folder = []
        if os.path.isdir(path):
            album_folder = [f for f in os.listdir(path) if os.path.isfile(path+'/'+f)]
        # self.cached_dirs[path] = album_folder
        return album_folder

    @staticmethod
    def same_file(f1, f2):
        with open(f1, 'rb') as f:
            c1 = f.read()
        with open(f2, 'rb') as f:
            c2 = f.read()
        return c1 == c2

    @staticmethod
    def rename_allowed(_stage, config):
        if config['enable-rename'] is False:
            return False
        return True

    @staticmethod
    def build_auto_config(src, dst, base):
        # we want to change the {base} component of names to base
        # because this value cannot be guessed from the directory name in ALBUM
        with open(src, 'r') as s:
            lines = s.readlines()
        with open(dst, 'w') as d:
            d.write(f"# file generated by --enable-auto-configuration on {datetime.datetime.now()}\n")
            for li in lines:
                m = re.match(r"names\s*:(.*)", li)
                if m:
                    li = li.replace('{base}', base)
                d.write(li)

    def run_stage_file(self, stage: str, config: PwpConfig, common_path: str, old_filename: str, base: str):
        LOGGER.incr_picture()

        LOGGER.debug('')
        LOGGER.debug(f"{stage} file path='{common_path}' old_filename='{old_filename}' base='{base}'")
        summary = f'{stage:6} {old_filename:31}: '  # will hold a 1 line summary of actions taken
        suffix = pathlib.Path(old_filename).suffix[1:]

        path = config[stage] + '/' + common_path
        file_path = path + "/" + old_filename
        backup = self.build_path(config['backup'], common_path, old_filename)

        if old_filename == 'Thumbs' or suffix in ["pl", "py", "log", "ini"]:   # FIXME: why Thumbs ???
            if stage == 'triage':
                LOGGER.info(f"backup {old_filename} to {backup}")
                LOGGER.msg(f"{summary} file type not managed COPY TO '{backup}'")
                ACTOR.copy(file_path, backup)
            return

        if suffix in ("bak", "md5"):
            if stage == 'triage':
                ACTOR.copy(file_path, backup)
            # silently avoid
            return

        move_to_album = (stage == 'triage')
        # if True, we need to move the file to album
        # in album stage, we move again to album only if the fle has been modified
        copy_file_path = config['tmp_dir'] + '/' + common_path + '/' + old_filename

        if suffix in ("mp4", "MP4", "txt", "TXT"):
            author = config.default_author()
            ACTOR.copy(file_path, copy_file_path, forced=True)

            summary += f"author({author+')':18} "

            suffix = suffix.lower()
            if suffix == "mp4":
                copy_object = PwpMp4(copy_file_path)
                new_date = copy_object.creation or PwpJpg.guess_file_date(file_path, config)
            else:
                copy_object = PwpObject(copy_file_path)
                new_date = PwpJpg.guess_file_date(file_path, config)

            summary += f"date({new_date}) "

        elif suffix not in ("jpg", "jpeg", "JPG", "JPEG"):
            LOGGER.msg(f"{summary} file type *** UNKNOWN *** : COPY TO '{backup}' ")
            ACTOR.copy(old_filename, backup)
            return
        else:
            suffix = "jpg"
            jpg = PwpJpg(file_path)
            new_date, author = jpg.get_jpg_new_date(config)
            summary += f"author({author+')':18} date({new_date}) "
            jpg.close()
            ACTOR.copy(file_path, copy_file_path, forced=True)
            copy_object = PwpJpg(copy_file_path)

        # here, the original file has been copied to a temporary file, copy_file_path
        # copy_file_path has not been modified.

        if copy_object.rotate(stage, config):
            if config['dryrun']:
                LOGGER.info(f"Would rotate/flip '{file_path}'")
            else:
                LOGGER.info(f"rotate/flip '{file_path}'")
            move_to_album = True  # picture was modified

            summary += "rot "
        else:
            summary += "    "
        if copy_object.set_metadata(stage, config, date=new_date, new_author=author):
            if config['dryrun']:
                LOGGER.info(f"Would insert metadata in '{file_path}'")
            else:
                LOGGER.info(f"metadata inserted in '{file_path}'")
            summary += "meta "
            move_to_album = True  # picture was modified
        else:
            summary += "     "
        copy_object.close()

        file_format = config.format('names')
        file_dico = config.format_dict(new_date, author, base=base, suffix=suffix)

        if self.rename_allowed(stage, config):
            new_filepath = file_format.format(**file_dico)
        else:
            # we do not change the name
            new_filepath = common_path + '/' + os.path.basename(file_path)

        # CAVEAT: in ALBUM, the path between ALBUM and the picture is reset by file_format !!!
        # so, common_path is NOT part of new_filepath !

        target = config['album'] + '/' + new_filepath
        target_rel_path = os.path.dirname(new_filepath)
        new_filename = os.path.basename(new_filepath)
        if suffix == 'txt':
            new_filename = old_filename
            new_filepath = target_rel_path + '/' + new_filename
            target = config['album'] + '/' + new_filepath

        to_increment = ACTOR.get_last_numerical_field(config['names'])
        album_folder = self.get_cached_folder(os.path.dirname(target))

        if new_filename not in album_folder:
            move_to_album = True  # it is not here, we need to put it

        if os.path.abspath(target) != os.path.abspath(file_path):
            move_to_album = True

        while new_filename in album_folder:
            if self.same_file(copy_file_path, target):
                if target == file_path:
                    LOGGER.info(f"File '{file_path}' has not changed")
                else:
                    LOGGER.info(f"New file '{file_path}' is already in album as '{target}'")
                move_to_album = False
                break
            elif target == file_path:
                # the file has been modified, so we need to clobber the existing file
                # message about modification was already done
                LOGGER.debug(f"Update '{file_path}' due to modifications")
                move_to_album = True
                break
            elif self.rename_allowed(stage, config):
                file_dico[to_increment] += 1
                new_filepath = file_format.format(**file_dico)
                target = config['album'] + '/' + new_filepath
                new_filename = os.path.basename(new_filepath)

                if suffix == 'txt':
                    new_filename = old_filename
                    new_filepath = target_rel_path + '/' + new_filename
                    target = config['album'] + '/' + new_filepath
                move_to_album = True
            else:
                LOGGER.debug(f"Clobber '{file_path}' with '{old_filename}' because rename not allowed")
                move_to_album = True
                break

        summary += f"-->({target}) "

        target_dir = os.path.dirname(target)
        if move_to_album:
            if stage == 'album':
                LOGGER.info(f"backup {old_filename} to {backup}")
                ACTOR.copy(file_path, backup)
                summary += f"backup({backup}) "
                ACTOR.delete(file_path)  # if dryrun, does nothing :-)

            if config['dryrun']:
                if file_path == target:
                    LOGGER.info(f"Would update '{file_path}'")
                else:
                    LOGGER.info(f"Would rename '{file_path}' : '{target}'")
                    # otherwise message is misleading: in reality, we do not rename to itself,
                    # we rename the copy that has been changed
                #  BUG:  target = file_path
            else:
                ACTOR.move(copy_file_path, target)
                if file_path == target:
                    LOGGER.info(f"Update '{file_path}'")
                else:
                    LOGGER.info(f"RENAME: '{file_path}' : '{target}'")

            # Put in cache if it was not there
            album_folder.append(new_filename)
            # self.cached_dirs[target_dir] = album_folder # in case album_folder was empty

            # here, target is always in ALBUM

            if config['enable-piwigo-sync'] and target_rel_path not in self.dirs_to_synchronize:
                self.dirs_to_synchronize.append(target_rel_path)  # for piwigo

        # even if the picture is unchanged, maybe the thumbs are not done

        if suffix == "jpg":
            if config['dryrun']:
                # we have to cheat, because target is NOT build !
                summary += self.build_all_thumbs(file_path, config, target_rel_path, stage, move_to_album)
            else:
                summary += self.build_all_thumbs(target, config, target_rel_path, stage, move_to_album)

        if not move_to_album and copy_file_path != file_path:
            ACTOR.delete(copy_file_path, forced=True)

        if stage != 'triage':
            LOGGER.msg(summary)
            return

        if config['enable-auto-configuration'] is False:
            LOGGER.msg(summary)
            return

        this_ini = path + '/' + config['ini-file']
        album_ini = target_dir + '/' + config['ini-file']

        if suffix == "jpg" and os.path.isfile(this_ini) and not os.path.isfile(album_ini):
            if config['dryrun']:
                LOGGER.info(f"Would Auto-configure '{this_ini}' to '{album_ini}'")
            else:
                LOGGER.info(f"Auto-configure '{this_ini}' to '{album_ini}'")
                self.build_auto_config(this_ini, album_ini, base)
            summary += "autoconf() "
        LOGGER.msg(summary)

    def verify_thumbnails(self, config: PwpConfig, common_path: str, old_filename: str):
        # path = config['album'] + '/' + common_path
        if old_filename[-4:] != '.jpg':
            return
        file_path = common_path + "/" + old_filename
        target_rel_path = os.path.dirname(file_path)
        target = config['album'] + '/' + file_path
        self.build_all_thumbs(target, config, target_rel_path, 'album', False)

    @staticmethod
    def synchronize(path, p_config):
        if p_config['dryrun'] or p_config['enable-piwigo-sync'] is False:
            return
        LOGGER.msg(f"synchronize '{path}'")
        if path[0] == '/':
            path = path[1:]
        all_dirs = os.path.dirname(path).split('/')

        done = ''
        for item in all_dirs:
            ACTOR.synchronize_piwigo_dir(done + item, with_files=False)
            done = done + item + '/'
        ACTOR.synchronize_piwigo_dir(path, with_files=True)

    # @staticmethod
    # def is_album_to_process(common_path, verify_albums):
    #     """
    #     test if the album needs to be processed
    #     :param common_path:
    #     :param verify_albums: True, False, or a list
    #     :return: bool
    #     """
    #     if isinstance(verify_albums, bool):
    #         raise PwpConfigError('verify-albums must be a list, not a bool')
    #
    #     if True in verify_albums:
    #         raise PwpConfigError('true is illegal in verify-albums')
    #     if False in verify_albums:
    #         raise PwpConfigError('false is illegal in verify-albums')
    #
    #     for d in verify_albums:
    #         if d.startswith(common_path):
    #             return True
    #     return False

    # @staticmethod
    # def allow_process_files_in_album(common_path, verify_albums):
    #     if isinstance(verify_albums, bool):
    #         raise PwpConfigError('verify-albums must be a list, not a bool')
    #
    #     if True in verify_albums:
    #         raise PwpConfigError('true is illegal in verify-albums')
    #     if False in verify_albums:
    #         raise PwpConfigError('false is illegal in verify-albums')
    #
    #     for d in verify_albums:
    #         if d == common_path:
    #             return True
    #     return False

    def remove_useless_thumbnails(self, path, config):
        if config['enable-thumbnails'] is False:
            LOGGER.debug(f" {path} enable-thumbnails is False")
            return
        if config['useless-thumbnails-delete'] is False:
            LOGGER.debug(f" {path} useless-thumbnails-delete is False")
            return

        thumbs_base = config['web'] + '/' + path
        if not os.path.isdir(thumbs_base):
            # this can happen, if there are no pictures in the corresponding album
            return

        all_files = os.listdir(thumbs_base)
        for file in all_files:
            file_path = thumbs_base + '/' + file
            if os.path.isfile(file_path):
                if file_path in self.expected_thumbnails:
                    pass  # this is normal
                elif config['dryrun']:
                    LOGGER.msg(f"SHOULD REMOVE {file_path}")
                else:
                    LOGGER.info(f"Removing extra thumbnail {file_path}")
                    ACTOR.delete(file_path)

        return  # because the debugger sometimes has issues with tail return

    def run_stage_dir(self, stage: str, p_config: PwpConfig, common_path: str, base: str, recursive=True):
        """
        stage: triages or albums
        common_path: the path, starting from TRIAGE, of the directory to run,
            it will be added to other roots
        base: the part of renamed files inherited from common_path
        """

        cur_dir = os.path.basename(common_path)

        if re.match(r'(.picasaoriginals.*|.idea.*|@eaDir)', cur_dir):
            # nothing to do here, this directory should not be managed
            return

        if stage == 'triage' and p_config['triage'] is None:
            LOGGER.msg("No TRIAGE directory")
            return

        if (stage == 'album' or stage == 'thumbnails') and p_config['album'] is None:
            LOGGER.msg("No ALBUM directory")
            return

        if stage == "thumbnails":
            path = p_config['album'] + ('' if common_path == '' else '/' + common_path)
        else:
            path = p_config[stage] + ('' if common_path == '' else '/' + common_path)

        new_conf_file = path + '/' + os.path.basename(p_config['ini-file'])
        new_conf = p_config.push_local_ini(new_conf_file)

        # if stage == 'album':
        #     if not self.is_album_to_process(common_path, p_config['verify-albums']):
        #         return

        LOGGER.msg('')
        LOGGER.msg(f"------ {stage} dir: common_path='{common_path}' base='{base}'")

        self.expected_thumbnails = []

        new_base = self.get_base_from_dir(cur_dir, p_config)

        if new_base and new_base != "":
            base = new_base

        all_files = os.listdir(path) if os.path.isdir(path) else []
        # self.cached_dirs = {}
        if recursive:
            for item in all_files:
                if os.path.isdir(path + '/' + item):
                    new_dir = common_path + ('' if common_path == '' else '/') + item
                    self.run_stage_dir(stage, new_conf, new_dir, base)

        self.expected_thumbnails = []

        for item in all_files:
            if os.path.isfile(path + '/' + item):
                # if stage == 'album' and not self.allow_process_files_in_album(common_path, p_config['verify-albums']):
                #     LOGGER.debug("Skipping file in album not leaf", common_path)
                # else:
                #     self.run_stage_file(stage, new_conf, common_path, item, base)
                if stage == 'thumbnails':
                    self.verify_thumbnails(new_conf, common_path, item)
                else:
                    self.run_stage_file(stage, new_conf, common_path, item, base)

        if stage == 'album' or stage == "thumbnails":
            self.remove_useless_thumbnails(common_path, p_config)

        if common_path == '':
            # we are back in the initial directory
            for path in self.dirs_to_synchronize:
                self.synchronize(path, p_config)

    def combined_config(self, path: str, p_config, prologue="", caller="dump_config"):
        # prologue is what we have already traversed, must end with '/' if not empty
        # path MUST be a relative path
        # the code here does NOT manage absolute paths

        if prologue == '':
            abs_path = os.path.abspath(path)
            if p_config["triage"] and abs_path.startswith(os.path.abspath(p_config["triage"])):
                rel_path = os.path.relpath(abs_path, os.path.abspath(p_config["triage"]))

                new_conf_file = p_config["triage"] + '/' + os.path.basename(p_config['ini-file'])
                new_conf = p_config.push_local_ini(new_conf_file)

                return self.combined_config(rel_path, new_conf, prologue=p_config["triage"] + '/', caller=caller)

            if p_config["album"] and abs_path.startswith(os.path.abspath(p_config["album"])):
                rel_path = os.path.relpath(abs_path, os.path.abspath(p_config["album"]))
                new_conf_file = p_config["album"] + '/' + os.path.basename(p_config['ini-file'])
                new_conf = p_config.push_local_ini(new_conf_file)
                return self.combined_config(rel_path, new_conf, prologue=p_config["album"] + '/', caller=caller)

            raise PwpConfigError(f"argument of {caller} must exist must be a subdirectory of triage or album",
                                 path)

        if path == '':
            self.dumped_config = p_config  # just for test
            return p_config

        if path[-1] != '/':
            path += '/'

        first = path.split('/')[0]
        if os.path.isdir(prologue + first):
            new_conf_file = prologue + first + '/' + os.path.basename(p_config['ini-file'])
            new_conf = p_config.push_local_ini(new_conf_file)
            new_path = path.replace(first + '/', '')
            new_prologue = prologue + '/' + first + '/' if prologue != '' else first + '/'
            return self.combined_config(new_path, new_conf, prologue=new_prologue, caller=caller)
        raise PwpConfigError(f"argument of -{caller} must exist", path)

    def dump_config(self, path, p_config):
        conf = self.combined_config(path, p_config, caller="dump_config")
        LOGGER.info(f"dump_config({path}): {conf['ini-filename-parsed']}")
        pprint.pprint(conf)

    def reset_ini(self):
        ini_file = self.parser_config['ini-file']
        if os.path.isfile(ini_file):
            ACTOR.move(ini_file, ini_file+'.bak')
        self.parser.build_ini_file(ini_file)
        LOGGER.msg(f"Reset '{ini_file}' to default")

    def run(self):
        do_exit = False  # exit is delayed until we have managed all --flag that generate exit

        # We do it HERE because we want to have read the 1st .ini in cwd

        ACTOR.configure(self.parser_config)
        LOGGER.configure(self.parser_config)

        # done automatically by argparse
        # if self.config['help']:
        #     self.parser.print_help()
        #     do_exit = True

        if self.parser_config['version']:
            version = PwpVersion()
            LOGGER.msg(f"current version: PyxlSql '{version.help}' ")
            do_exit = True

        if self.parser_config['licence']:
            licence = PwpLicence()
            licence.print()
            do_exit = True

        # -------------------------------------------------------------
        # SSH CONNECT !
        #
        # has already been done in ACTOR.configure

        remote, uname, cause, is_error = ACTOR.connect_ssh(self.parser_config)

        if self.parser_config['test-ssh']:
            LOGGER.msg("Testing ssh ")
            LOGGER.msg(f"remote host : '{ACTOR.remote_host}'")
            LOGGER.msg(f"remote port : '{ACTOR.remote_port}'")
            LOGGER.msg(f"remote user : '{ACTOR.remote_user}'")
            LOGGER.msg(f"uname       : '{uname}'")
            if remote:
                result = ACTOR.remote_ls(".")
                LOGGER.info(f"test-ssh OK: ls -l       : '{result}'")
            else:
                LOGGER.info(f"Cannot ssh : {cause}")
            do_exit = True

        if self.parser_config['test-sftp']:
            if not remote:
                LOGGER.msg("sftp test OK : skipped because no remote configuration")
                LOGGER.info("sftp test OK : skipped because no remote configuration")
            else:
                dummy = "dummy.txt"
                dummy_timestamp = ACTOR.build_timestamp(dummy)
                base = os.path.basename(dummy)
                LOGGER.debug("Testing sftp")

                dst = self.parser_config['remote-web']
                ACTOR.remote_put(dummy, dst)
                result = ACTOR.remote_ls(dst)
                if base not in result:
                    LOGGER.info(f"sftp failed      : '{result}'")
                    LOGGER.msg("sftp test failed")
                else:
                    remote_stamp = ACTOR.timestamp_from_ls(result[base])
                    if remote_stamp == dummy_timestamp:
                        LOGGER.info(f"sftp test OK          : '{result[base]}'")
                        LOGGER.msg("sftp test OK")
                    else:
                        LOGGER.info(f"sftp set time failed      : '{dummy_timestamp}'  '{remote_stamp}")
                        LOGGER.msg("sftp test failed")

                ACTOR.remote_rm(dst + '/' + dummy)
            do_exit = True

        src = self.parser_config['test-piwigo-synchronization']
        if src is not None:
            LOGGER.msg(f"Testing piwigo synchronization of '{src}'")
            ACTOR.synchronize_piwigo_dir(src, with_files=False)
            do_exit = True

        if self.parser_config['reset-ini']:
            self.reset_ini()
            do_exit = True

        if self.parser_config['dump-config']:
            self.dump_config(self.parser_config['dump-config'], self.parser_config)
            do_exit = True

        if do_exit:
            LOGGER.msg("Exiting due to cmdline options")
        else:
            with tempfile.TemporaryDirectory() as tmp_dir:
                self.parser_config['tmp_dir'] = tmp_dir

                if self.parser_config['verify-albums']:
                    if self.parser_config['triage']:
                        self.parser_config['triage'] = None
                        LOGGER.warning(f"removing target --triage {self.parser_config['triage']} " +
                                       "because --verify-album not empty")

                    if self.parser_config['verify-thumbnails']:
                        self.parser_config['verify-thumbnails'] = None
                        LOGGER.warning("removing target --verify-thumbnails " +
                                       f"{self.parser_config['verify-thumbnails']} " +
                                       "because --verify-album not empty")

                    for item in self.parser_config['verify-albums']:
                        item_father = os.path.dirname(item)
                        common_path = self.parser_config['album'] + ('/' + item_father if item_father else "")
                        cur_dir = os.path.basename(item)
                        p_config = self.combined_config(common_path, self.parser_config, caller="verify-albums")
                        new_base = self.get_base_from_dir(cur_dir, p_config)
                        self.run_stage_dir('album', p_config, item, new_base, recursive=False)

                elif self.parser_config['verify-thumbnails']:
                    if self.parser_config['triage']:
                        self.parser_config['triage'] = None
                        LOGGER.warning(f"removing target --triage {self.parser_config['triage']} " +
                                       "because --verify-thumbnails not empty")

                    for item in self.parser_config['verify-thumbnails']:
                        item_father = os.path.dirname(item)
                        common_path = self.parser_config['album'] + ('/' + item_father if item_father else "")
                        cur_dir = os.path.basename(item)
                        p_config = self.combined_config(common_path, self.parser_config)
                        new_base = self.get_base_from_dir(cur_dir, p_config)
                        self.run_stage_dir('thumbnails', p_config, item, new_base, recursive=True)
                else:
                    self.run_stage_dir('triage', self.parser_config, '', '', recursive=True)

            LOGGER.msg("End of processing")

        LOGGER.end()
        os.chdir(self.initial_cwd)


def pwp_init(arguments=None):
    """used for tests, when the test harness in test_pwp needs to use the ssh connection
    initializes PwpMain"""

    main = PwpMain(arguments)
    return main


def pwp_run(main: PwpMain):
    """used for tests, when the test harness in test_pwp needs to use the ssh connection"""
    if main is None:
        raise PwpError("main is None")
    main.run()
    return main.parser_config if main is not None else None


def pwp_main(arguments=None):
    main = PwpMain(arguments)
    if main.parser_config is None:
        return None
    return pwp_run(main)


if __name__ == "__main__":
    pwp_main(sys.argv[1:])



# see https://www.kaggle.com/code/humananalog/examine-mp4-files-with-python-only

# More information about the MP4 file format at the following links:
#
#     http://xhelmboyx.tripod.com/formats/mp4-layout.txt
#     https://github.com/OpenAnsible/rust-mp4/raw/master/docs/ISO_IEC_14496-14_2003-11-15.pdf
#     https://developer.apple.com/library/archive/documentation/QuickTime/QTFF/QTFFPreface/qtffPreface.html

# import os, sys
import struct
import datetime
import sys
import argparse

from piwiPre.pwpActor import ACTOR
from piwiPre.pwpErrors import PwpError, LOGGER
from piwiPre.pwpJpg import PwpObject


class PwpMp4Box:
    having_sons = ["moov", "udta", "trak"]  # noqa

    def __init__(self, f, father, context, offset: int, get_boxes=True):
        self.option_1 = False
        self.length = None
        self.name = None
        self.data = None
        self.father = father
        self.context = context
        self.offset = offset
        self.file = f
        self.header_length = 8
        self.boxes = {}
        self.boxes_list = []

        self.length = self.get_unsigned(f, 4)
        if self.length is None:
            return

        self.name = self.get_str(f, 4)
        # Note: if any box grows in excess of 2^32 bytes (> 4.2 GB), the box size can be extended
        # in increments of 64 bits (18.4 EB).
        # By setting the box size to 1 and appending a new 64 bit box size.
        if self.length == 1:
            self.option_1 = True
            data = f.read(8)
            self.length, = struct.unpack("> Q", data)
            self.header_length += 8
        if get_boxes and self.name in self.having_sons:
            self.boxes, self.boxes_list = PwpMp4.find_boxes(f, self)
            f.seek(self.offset + self.length, 0)  # because some boxes may have been read not completely
        else:
            self.data = f.read(self.length - self.header_length)

    def pack_header(self):
        s = struct.Struct("> I 4s")
        s1 = struct.Struct("> I 4s Q")
        if self.option_1:
            data = s1.pack(1, self.name.encode(), self.length)
        else:
            data = s.pack(self.length, self.name.encode())
        return data

    def print(self, suffix=""):
        fl = " Header type 1" if self.option_1 else ""
        fn = self.father.name if self.father else "root"
        LOGGER.msg(f"{suffix}box({fn}):'{self.name}' offset {self.offset:8} length {self.length:8} {fl} {self.context}")
        for box in self.boxes_list:
            box.print(suffix=suffix + "  ")
        if len(self.boxes_list) > 0:
            LOGGER.msg(f"{suffix}--- end of {self.name}")

    @staticmethod
    def get_str(f, length):
        data = f.read(length)
        if data == b"":
            return None
        val = data.decode('utf8')
        return val

    @staticmethod
    def get_unsigned(f, length):
        data = f.read(length)
        if data == b"":
            return None
        val = int.from_bytes(data, "big")
        return val

    def equals(self, box):
        return self.name == box.name and self.offset == box.offset and \
            self.option_1 == box.option_1 and self.length == box.length


class PwpMp4(PwpObject):
    def __init__(self, filename):
        super().__init__(filename)

        self.root_boxes = {}
        self.root_list = []
        self.moov_boxes = {}  # noqa
        self.moov_list = []  # noqa
        self.creation = None
        self.copyright = None
        self.udta_boxes = {}  # noqa
        self.udta_list = []  # noqa
        self.cprt = None  # noqa

        with open(filename, 'rb') as f:
            self.root_boxes, self.root_list = self.find_boxes(f, None)

            if "ftyp" not in self.root_boxes or self.root_boxes["ftyp"].offset != 0:  # noqa
                raise PwpError("wrong MP4 structure: offset(ftyp) not 0", filename)  # noqa

            if "moov" not in self.root_boxes:  # noqa
                raise PwpError("wrong MP4 structure: no moov box ", filename)  # noqa

            self.moov_boxes, self.moov_list = self.find_boxes(f, self.root_boxes["moov"])  # noqa

            self.creation = self.get_creation(f, self.moov_boxes["mvhd"])  # noqa

            if "udta" in self.moov_boxes:  # noqa
                self.udta_boxes, self.udta_list = self.find_boxes(f, self.moov_boxes["udta"])  # noqa
                if "cprt" in self.udta_boxes:  # noqa
                    self.cprt = self.udta_boxes["cprt"]  # noqa
                    self.copyright = self.get_copyright(f, self.cprt)

    def print(self):
        LOGGER.msg(f"mp4       '{self.filename}'")
        LOGGER.msg(f"creation   {self.creation}")
        LOGGER.msg(f"Copyright '{self.copyright}'")
        for box in self.root_list:
            box.print("   ")

    @staticmethod
    def get_copyright(f, box: PwpMp4Box):
        f.seek(box.offset + box.header_length + 6, 0)  # skip header, lang
        char = 1
        res = bytes()
        while char != b'\x00':
            char = f.read(1)
            if char != b'\x00':
                res += char
        res_str = res.decode('utf8')
        return res_str

    @staticmethod
    def find_boxes(f, father: PwpMp4Box | None):
        """Returns a dictionary of all the data boxes and their absolute starting
        and ending offsets inside the mp4 file.

        when finished, the seek pointer is at end of data
        """
        if father is None:
            offset = 0
            end_offset = float("inf")
        else:
            offset = father.offset + father.header_length
            end_offset = father.offset + father.length

        boxes = {}
        boxes_list = []

        f.seek(offset, 0)
        while offset < end_offset - 8:
            box = PwpMp4Box(f, father, "parsing", offset)
            if box.length is None:
                return boxes, boxes_list
            # box.print()
            boxes[box.name] = box
            boxes_list.append(box)
            offset += box.length
        return boxes, boxes_list

    def add_box(self, filename, added_data: []):
        """Adds a box at the end of the current box described by the offset"""
        added_length = len(added_data)
        current = self.root_boxes["moov"]  # noqa
        udta_length = 0 if self.copyright is None else self.moov_boxes["udta"].length  # noqa

        insert = True
        with open(filename, "wb") as nf, open(self.filename, 'rb') as f:
            f_offset = 0
            n_offset = 0

            data = f.read(current.offset)
            f_offset += current.offset

            nf.write(data)
            n_offset += current.offset
            assert n_offset == f_offset  # noqa before the header of moov

            moov = PwpMp4Box(f, None, "Adding a box", current.offset, get_boxes=False)  # noqa
            f_offset += moov.length  # noqa after the moov box, because we read the entire box
            # noqa moov.print()
            assert current.equals(moov)

            if insert:
                moov.length += added_length - udta_length
                # noqa print(f"new length of moov {moov.length}")
            data = moov.pack_header()
            nf.write(data)
            n_offset += len(data)
            nf.write(moov.data)
            n_offset += len(moov.data)
            if insert and udta_length:
                nf.seek(-udta_length, 1)  # noqa go back to the start of udta
                n_offset -= udta_length
            assert n_offset == f_offset - udta_length  # noqa udta should be inserted

            # the new box must be inserted in the padding area reserved in the next box which must be "free"
            # and we lower the free zone for the same amount.

            if insert:
                nf.write(added_data)
                n_offset += len(added_data)
                assert n_offset == f_offset + (added_length - udta_length)  # noqa after udta

            # follow box should be 'free'
            free = PwpMp4Box(f, None, "next insert", f_offset, get_boxes=False)
            assert free.name == 'free'
            # free.print()
            initial_free_length = free.length

            if insert:
                free.offset = f_offset + (added_length - udta_length)
                free.length -= (added_length - udta_length)
                free.context = "modified for insert"
                # free.print()

            free_header = free.pack_header()
            nf.write(free_header)
            n_offset += len(free_header)
            assert len(free_header) == free.header_length

            assert n_offset == f_offset + (added_length - udta_length) + len(free_header)  # after the new free header

            f_offset += initial_free_length  # the real position of the seek index

            nf.write(bytes(1) * (free.length - free.header_length))
            # here free.length is the real length after removing added_length
            n_offset += (free.length - free.header_length)

            assert n_offset == f_offset

            # we should find here new boxes
            while True:
                mdat = PwpMp4Box(f, None, "after insert", f_offset, get_boxes=False)
                if mdat.length is None:
                    break  # EOF
                # free.print()

                data = mdat.pack_header()
                nf.write(data)
                nf.write(mdat.data)
                f_offset += mdat.length + mdat.header_length - 8
            return

    @staticmethod
    def get_creation(f, box: PwpMp4Box):

        f.seek(box.offset, 0)
        new_h = PwpMp4Box(f, None, "Getting creation date", box.offset, get_boxes=False)
        assert box.equals(new_h)

        f.seek(box.offset + box.header_length, 0)
        version = PwpMp4Box.get_unsigned(f, 1)
        word_size = 8 if version == 1 else 4

        # 3 bytes flags =  24-bit hex flags (current = 0)
        f.seek(3, 1)  # skip flags
        # 4 bytes created mac UTC date
        #          = long unsigned value in seconds since beginning 1904 to 2040
        created = PwpMp4Box.get_unsigned(f, word_size)
        utc_time = datetime.datetime(1904, 1, 1) + datetime.timedelta(seconds=created)
        return utc_time

    # noqa  * 8+ bytes optional user data (any custom info) box
    # noqa         = long unsigned offset + long ASCII text string 'udta'
    #
    # noqa        * 8+ bytes optional copyright notice box
    # noqa            = long unsigned offset + long ASCII text string 'cprt'
    # noqa          -> 4 bytes version/flags = byte hex version + 24-bit hex flags
    # noqa              (current = 0)
    # noqa          -> 1/8 byte ISO language padding = 1-bit value set to 0
    # noqa          -> 1 7/8 bytes content language = 3 * 5-bits ISO 639-2 language code less 0x60
    # noqa               ==> 1 + 15 bits = 2 bytes.
    # noqa            - example code for english = 0x15C7 qui fait très exactement 'eng'
    # noqa          -> annotation string = UTF text string
    # noqa          -> 1 byte annotation c string end = byte value set to zero
    # noqa        -> 4 bytes compatibility utda end = long value set to zero

    @staticmethod
    def build(annotation: str):  # noqa
        anno_len = len(annotation)
        # noqa                udta_length   cprt_length
        length = anno_len + 27  # 4
        descr = b'udta'  # noqa 4
        length_2 = anno_len + 15  # 4
        descr_2 = b'cprt'  # noqa          4
        version = 0  # 4
        lang: int = 0x15C7  # 2
        # annotation        #               l + 1
        com = 0  # 4
        # noqa                 Total cprt # l + 15
        # noqa  total udta  # l + 27

        s = struct.Struct(f"> I 4s I 4s I H {anno_len + 1}s I")
        data = s.pack(length, descr, length_2, descr_2, version, lang, annotation.encode(), com)
        assert len(data) == length
        # noqa print(f"udta length = {length} {len(data)}")
        return data

    def write_metadata(self, _new_author, copy_right, _new_special):
        copy_name = self.filename[:-4] + "_bak.mp4"
        ACTOR.copy(self.filename, copy_name)
        son = PwpMp4(copy_name)
        if son.copyright == copy_right:
            return False
        udta2 = self.build(copy_right)  # noqa
        son.add_box(self.filename, udta2)  # noqa
        return True


def diff(fn1, fn2):
    with open(fn1, 'rb') as f1, open(fn2, 'rb') as f2:
        index = 0
        u1 = 0
        u2 = 0
        print(f"comparing {fn1} and {fn2}")
        while u1 is not None and u2 is not None:
            if u1 != u2:
                print(f"{index:8}: {u1:08x} {u2:08x}")
            u1 = PwpMp4Box.get_unsigned(f1, 4)
            u2 = PwpMp4Box.get_unsigned(f2, 4)
            index += 4
        if u1 is not None:
            print(f"f1 is longer {index}")
        elif u2 is not None:
            print(f"f2 is longer {index}")
        else:
            print("done")


def mp4_main(arguments):
    parser = argparse.ArgumentParser(description='test copyright insertion in MP4 file')
    parser.add_argument('--file', '-f', help="source file",
                        action='store',
                        default="tests/sources/TRIAGE3/Vendée/Opéra-Paris.mp4")
    parser.add_argument('--output', '-o', help="output",
                        action='store',
                        default="tests/results/copyrighted.mp4")
    parser.add_argument('--diff', '-d', help="compare the file",
                        action='store_true')
    parser.add_argument('--print', '-p', help="print files",
                        action='store_true')

    args = parser.parse_args() if arguments is None else parser.parse_args(arguments)
    mp4 = PwpMp4(args.file)
    if args.print:
        mp4.print()
    if "moov" not in mp4.root_boxes is None:  # noqa
        raise PwpError("No moov box, exit", args.file)  # noqa

    c1 = "Copyright 2013 foo BAR"
    udta = mp4.build(c1)  # noqa
    mp4.add_box(args.output, udta)  # noqa

    son = PwpMp4(args.output)
    if son.copyright != c1:
        raise PwpError(f"wrong copyright '{c1}' != '{son.copyright}'")

    if args.print:
        son.print()
    if args.diff:
        diff(args.file, args.output)

    n2 = args.output[:-4] + "-v2.mp4"
    c2 = "Copyright 2013 lee-foo BAR-BAR"
    udta2 = son.build(c2)  # noqa
    son.add_box(n2, udta2)  # noqa

    son2 = PwpMp4(n2)
    if son2.copyright != c2:
        raise PwpError(f"wrong copyright '{c2}' != '{son2.copyright}'")

    if args.print:
        son2.print()
    if args.diff:
        diff(args.output, n2)


if __name__ == "__main__":
    mp4_main(sys.argv[1:])

# cf https://gpac.github.io/mp4box.js/test/filereader.html



# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini(at)gmail.com
# ---------------------------------------------------------------------------------------------------------------


# ---------------------------------------------------------------------------------------------------------------
# test_pwp.py
# ---------------------------------------------------------------------------------------------------------------
# This test is intended to be run from piwiPre directory by invoking pytest,
# which will run all files of the form test_*.py or *_test.py in the current directory and its subdirectories.
#
# is called by tox
#
# can be called directly by running, from piwiPre root directory (where tox.ini is):
# prompt> python -I tests/test_pwp.py -n 0 # or other number...

# =============================================================================================================
# CAVEAT:
# =============================================================================================================
#         tox has issues with comodo antivirus when run from PyCharm Terminal window
#         To avoid this, run tox from a native terminal, not within PyCharm
#
#         The origin seems to be that PyCharm creates a temp script which is not assess correctly by comodo
# =============================================================================================================


import sys
import os
import argparse
import inspect
import re
import datetime

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from piwiPre.pwpActor import ACTOR
from piwiPre.pwpMain import pwp_main, pwp_run, pwp_init
from piwiPre.pwpErrors import PwpConfigError, LOGGER  # , PwpError
from piwiPre.pwpArgsIni import args_ini_main
from piwiPre.pwpPatcher import patcher_main
from piwiPre.pwpParser import pwp_parser_main
from piwiPre.pwpJpg import PwpJpg
from piwiPre.pwpMp4 import mp4_main, PwpMp4
# from piwiPre.pwpConfig import PwpConfig
from piwiPre.pwpParser import PwpParser

from source.parse_requirements import run_requirements


class PwpTester:
    def __init__(self):
        self.programs_done = {'program_0': False}
        self.programs = {'program_0': self.program_0}
        self.run_file = "tests/sources/run.txt"
        self.asserts = 0
        self.start_time = datetime.datetime.now()
        self.pictures_nb = 0

        parser = PwpParser(arguments=[], program="test_harness", with_config=True)
        self.home_config = parser.home_config

        # initial_config = parser.build_initial_config()
        # home_ini_path = os.path.expanduser("~") + '/.piwiPre.ini'
        # first_ini = PwpConfig.parse_ini_file(home_ini_path)
        # self.home_config = first_ini.merge_ini(initial_config)

        members = inspect.getmembers(self)
        last = 0
        for m in members:
            r = re.match(r"program_\d+", m[0])
            if r:
                name = m[0]
                number = int(name[8:])  # because we want them ordered
                last = max(number, last)
                self.programs_done[name] = False
                self.programs[name] = m[1]
        self.last = last

    @staticmethod
    def get_assert_context():
        previous_frame = inspect.currentframe().f_back.f_back
        (filename, line_number, _function_name, _lines, _index) = inspect.getframeinfo(previous_frame)
        context = f"'{filename}':{line_number:03}\n               "
        return context

    def assert_dir(self, dir_path: str):
        msg = f"{self.get_assert_context()} directory {dir_path} should exist"
        self.asserts += 1
        assert os.path.isdir(dir_path), msg

    def assert_no_dir(self, dir_path: str):
        msg = f"{self.get_assert_context()} directory {dir_path} should not exist"
        self.asserts += 1
        assert not os.path.isdir(dir_path), msg

    def assert_file(self, filepath: str):
        msg = f"{self.get_assert_context()} file {filepath} should exist"
        self.asserts += 1
        assert os.path.isfile(filepath), msg

    def assert_no_file(self, filepath: str):
        msg = f"{self.get_assert_context()} file {filepath} should not exist"
        self.asserts += 1
        assert not os.path.isfile(filepath), msg

    def assert_file_contains(self, filepath: str, line: str):
        msg = f"{self.get_assert_context()} file {filepath} should exist"
        self.asserts += 1
        assert os.path.isfile(filepath), msg
        with open(filepath, 'r', encoding="utf-8") as f:
            lines = f.readlines()
        msg = f"{self.get_assert_context()} file {filepath} should contain '{line}'"
        for la in lines:
            if line in la:
                return
        assert False, msg

    def assert_file_not_contains(self, filepath: str, line: str):
        msg = f"{self.get_assert_context()} file {filepath} should exist"
        self.asserts += 1
        assert os.path.isfile(filepath), msg
        with open(filepath, 'r', encoding="utf-8") as f:
            lines = f.readlines()
        msg = f"{self.get_assert_context()} file {filepath} should not contain '{line}'"
        for la in lines:
            assert line not in la, msg

    def assert_config(self, config, item, value):
        msg = f"{self.get_assert_context()} config[{item}] should be '{value}' and not '{str(config[item])}'"
        self.asserts += 1
        assert str(config[item]) == str(value), msg

    def assert_not_config(self, config, item, value):
        msg = f"{self.get_assert_context()} config[{item}] should not be '{value}'"
        self.asserts += 1
        assert str(config[item]) != str(value), msg

    def assert_error_number(self, level: str, err_nb: int):
        nb = LOGGER.msg_nb(level)
        msg = f"Line:{str(inspect.currentframe().f_lineno)}: {level}[{err_nb}] is not the max[{nb}]"
        self.asserts += 1
        assert nb == err_nb, msg

    def assert_error_contains(self, level: str, err_nb: int, start: str):
        nb = LOGGER.msg_nb(level)
        self.asserts += 1
        previous_frame = inspect.currentframe().f_back
        (filename, line_number, _function_name, _lines, _index) = inspect.getframeinfo(previous_frame)
        context = f"{filename}:{line_number:3}"
        msg = f"{context}: Error[{err_nb}] not reached"
        assert nb >= err_nb, msg
        if nb < err_nb:
            return
        err = LOGGER.data[level][err_nb]
        msg = f"{context}: Error[{err_nb}] '{err[:40]}' != '{start[:40]}' "
        assert start in err, msg

    def assert_info(self, value: str):
        msg = f"{self.get_assert_context()} debug info log should contain '{value}'"
        self.asserts += 1
        all_info = LOGGER.data['info']
        for err in all_info:
            if value in err:
                return
        assert False, msg

    def assert_info_or(self, values: [str]):
        msg = f"{self.get_assert_context()} debug info log should contain '{values[0]}' or ..."
        self.asserts += 1
        all_info = LOGGER.data['info']
        for err in all_info:
            for v in values:
                if v in err:
                    return True
        assert False, msg

    def assert_not_info(self, value: str):
        msg = f"{self.get_assert_context()} debug info log should contain '{value}'"
        self.asserts += 1
        for err in LOGGER.data['info']:
            assert value not in err, msg

    def assert_orientation(self, jpg: PwpJpg, rot: int):
        val = jpg.orientation
        msg = f"{self.get_assert_context()} orientation should be {rot} not {val}"
        self.asserts += 1
        assert val == rot, msg

    def assert_no_orientation(self, jpg: PwpJpg, rot: int):
        val = jpg.orientation
        msg = f"{self.get_assert_context()} orientation should not be {rot} "
        self.asserts += 1
        assert val != rot, msg

    def assert_copyright(self, jpg: PwpJpg, expected: str):
        val = jpg.copyright
        msg = f"{self.get_assert_context()} copyright '{val}' should contain '{expected}' "
        self.asserts += 1
        assert expected in str(val), msg

    def assert_author(self, jpg: PwpJpg, expected: str):
        val = jpg.author
        msg = f"{self.get_assert_context()} author '{val}' should contain '{expected}' "
        self.asserts += 1
        assert expected in str(val), msg

    def assert_mp4_copyright(self, mp4: PwpMp4, expected: str):
        val = mp4.copyright
        msg = f"{self.get_assert_context()} copyright '{val}' should contain '{expected}' "
        self.asserts += 1
        assert expected in str(val), msg

    def assert_no_copyright(self, jpg: PwpJpg, expected: str):
        val = jpg.copyright
        msg = f"{self.get_assert_context()} copyright '{val}' should not contain '{expected}' "
        self.asserts += 1
        assert expected not in str(val), msg

    def assert_special(self, jpg: PwpJpg, expected: str):
        val = jpg.special
        msg = f"{self.get_assert_context()} special instructions '{val}' should contain '{expected}' "
        self.asserts += 1
        assert expected in str(val), msg

    @staticmethod
    def create(filename):
        with open(filename, 'w') as f:
            f.write(f"Fake file created for test {datetime.datetime.now()}\n")

    def done(self, name):
        ACTOR.mkdirs("tests/results")
        self.create(f"tests/results/{name}_done.txt")

    @staticmethod
    def check_done(name: str):
        return os.path.isfile(f"tests/results/{name}_done.txt")

    @staticmethod
    def reset_done(name: str):
        ACTOR.delete(f"tests/results/{name}_done.txt", forced=True)

    def reset_data(self):
        ACTOR.reset_data()
        self.pictures_nb += LOGGER.reset_data()

    # --------------------------------------- Testing cmd line flags and configuration
    def program_1(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('')
        LOGGER.msg('testing chdir, reset-ini, dump-config')
        LOGGER.msg('')

        ACTOR.rmtree("tests/results")
        ACTOR.mkdirs("tests/results/TRIAGE/Armor")
        ACTOR.copy('tests/sources/piwiPre-fake.ini', "tests/results/piwiPre.ini")
        ACTOR.copy('tests/sources/piwiPre-alt-Armor.ini', "tests/results/TRIAGE/Armor/piwiPre.ini")

        pwp_main(['--reset-ini',  '--chdir', 'tests/results/TRIAGE'])
        self.assert_file('tests/results/TRIAGE/piwiPre.ini')  # the file is there, it should overwrite info in cwd

        LOGGER.msg('')
        LOGGER.msg('Step 2 in the same program, --chdir')
        LOGGER.msg('')

        main = pwp_init(['--chdir', 'tests/results', '--dump-config', 'TRIAGE/Armor'])  # , '--debug'])
        pwp_run(main)

        armor_conf = main.dumped_config
        # the hierarchy should be:
        # cwd (aka tests/results): piwiPre-fake.ini
        # cwd/TRIAGE:              --reset-ini
        # cwd/TRIAGE/Armor:        piwiPre-alt-Armor.ini

        # album is written by cwd/piwiPre.ini and reset in cwd/TRIAGE
        self.assert_config(armor_conf, 'album', 'ALBUM')

        # piwigo-user should be set in cwd/TRIAGE, but overwritten by cwd/TRIAGE/Armor
        self.assert_config(armor_conf, 'piwigo-user', 'unknown-user')
        self.assert_config(armor_conf, 'remote-user', 'None')

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_2(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')

        LOGGER.msg('testing --licence etc')
        pwp_main(['--chdir', 'tests/results', '--licence', '--version'])
        # PwpLicence.print()
        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        # verify visual output, mainly to get coverage higher
        self.done(mn)

    def program_3(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        ACTOR.mkdirs('tests/results')
        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('')
        LOGGER.msg('testing  --help')
        LOGGER.msg('')
        pwp_main(['--chdir', 'tests/results', '--help'])
        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_4(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')

        LOGGER.msg('')
        LOGGER.msg('testing --ini-file, all behaviors')
        LOGGER.msg('')
        ACTOR.rmtree("tests/results")
        ACTOR.copytree("tests/sources/TRIAGE2/Armor-cup", "tests/results/TRIAGE/Armor-cup")
        ACTOR.copy('tests/sources/piwiPre-alt.ini', "tests/results")
        ACTOR.copy('tests/sources/piwiPre-alt-Armor.ini', "tests/results/TRIAGE/Armor-cup/piwiPre-alt.ini")
        ACTOR.copy('tests/sources/piwiPre.ini', "tests/results")

        config = pwp_main(['--chdir', 'tests/results', '--ini-file', 'piwiPre-alt.ini'])  # , '--debug'])

        # let's verify the alternative .ini has been used, even in subdirectories.
        # the album and web changes are in cwd
        # the date change is in Armor-cup

        self.assert_config(config, 'album', 'LOCAL-ALBUM')
        self.assert_config(config, 'piwigo-user', 'Foo')
        self.assert_no_file('tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/' +  # noqa
                            '2023-06-11-17h02-26-Forêt-de-la-Corbière.jpg')  # noqa
        self.assert_no_file('tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/' +  # noqa
                            '2023-06-11-17h02-26-Forêt-de-la-Corbière.jpg')  # noqa

        LOGGER.msg(f"cwd  = '{os.getcwd()}' ")

        self.assert_dir('tests/results')  # noqa
        self.assert_dir('tests/results/LOCAL-ALBUM')  # noqa
        self.assert_dir('tests/results/LOCAL-ALBUM/2023')  # noqa
        self.assert_dir('tests/results/LOCAL-ALBUM/2023/2023-08-Aug-30-Armor-cup')  # noqa

        self.assert_file('tests/results/LOCAL-ALBUM/2023/2023-08-Aug-30-Armor-cup/' +  # noqa
                         '2023-08-30-11h05-44-Armor-cup.jpg')  # noqa
        self.assert_file('tests/results/LOCAL-ALBUM/2023/2023-08-Aug-30-Armor-cup/' +  # noqa
                         '2023-08-30-11h13-50-Armor-cup.jpg')  # noqa
        self.assert_file('tests/results/LOCAL-ALBUM/2023/2023-08-Aug-30-Armor-cup/' +  # noqa
                         '2023-08-30-11h31-28-Armor-cup.jpg')  # noqa
        self.assert_file('tests/results/LOCAL-WEB/2023/2023-08-Aug-30-Armor-cup/' +  # noqa
                         '2023-08-30-11h05-44-Armor-cup-cu_e250.jpg')  # noqa
        LOGGER.msg(f'--------------- end of  {mn}')
        self.done(mn)

    def program_10(self):
        initial_cwd = os.getcwd()
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('autotest of ArgsIni')
        os.chdir('tests/argsini')    # noqa
        args_ini_main(['--auto-test', '--verify-albums', 'fifi', '--verify-albums', 'toto'])
        os.chdir(initial_cwd)
        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_11(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('autotest of pwpPatcher')
        ACTOR.rmtree("tests/results")
        ACTOR.mkdirs("tests/results")
        # this test must be executed, like pwpPatcher, from piwPre root
        patcher_main(['--autotest'])
        self.assert_file_contains('tests/results/pwpLicence.py.autotest', 'EUROPEAN UNION PUBLIC LICENCE')
        self.assert_file_contains('tests/results/pwpVersion.py.autotest', 'class PwpVersion:')
        self.assert_file('tests/results/version.txt.autotest')
        self.assert_file_contains('tests/results/configuration.rst.autotest', '**enable-rotation** : true')
        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def get_first_album(self):
        LOGGER.msg('')
        LOGGER.msg('if run by the developer, will verify album "2023/2023-01-Janvier-08-Ballade"')  # noqa
        LOGGER.msg("if run in a different context, will run the first sub-album in album")
        LOGGER.msg("If no piwigo-host/user defined, will abort")
        LOGGER.msg('')

        if self.home_config['piwigo-host'] is None or self.home_config['piwigo-user'] is None:
            LOGGER.msg("No piwigo-host/user defined, aborting")
            return None
        album = self.home_config['album']
        if os.path.isdir(album):
            return "2023/2023-01-Janvier-08-Ballade"  # noqa

        if not os.path.isdir(album):
            return None

        all_files = os.listdir(album)
        for it in all_files:
            if os.path.isdir(album + '/' + it):
                return it
        return None

    def program_12(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('autotest of piwigo synchronisation')
        ACTOR.rmtree("tests/results")
        ACTOR.mkdirs("tests/results")
        # fallback to HOME for piwigo information

        src = self.get_first_album()
        if src:
            pwp_main(['--chdir', 'tests/results', '--test-piwigo-synchronization', src])
            self.assert_info(f"dir '{src}' synchronized with piwigo")
        else:
            LOGGER.msg("aborting test")

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')

        self.done(mn)

    def program_13(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('autotest of PwpParser')
        os.chdir('tests/argsini')    # noqa
        pwp_parser_main([])
        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        # self.done(mn) useless, and test/results does not exist

    def program_14(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('')
        LOGGER.msg('autotest of mp4 management')
        LOGGER.msg('')
        ACTOR.rmtree("tests/results")
        ACTOR.mkdirs("tests/results")
        mp4_main(['-p'])
        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_15(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('Error generation')
        seen = False
        try:
            pwp_main(['--chdir', '--ini-file', 'piwiPre-alt.ini'])
        except PwpConfigError:
            seen = True
        assert seen, "program_26 should have generated a PwpConfigError"

        LOGGER.msg(f'--------------- end of  {mn}')
        self.done(mn)

    def program_16(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('autotest of of parse_requirements')
        run_requirements([])
        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_17(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('Errors in parse_requirements')
        ACTOR.copytree("tests/sources/TRIAGE", "tests/results/TRIAGE")
        ACTOR.copy("tests/sources/piwiPre-error.ini", "tests/results/piwiPre.ini")
        initial_cwd = os.getcwd()

        done = False
        try:
            main = pwp_init(['--chdir', 'tests/results'])
            pwp_run(main)
        except PwpConfigError:
            done = True
            LOGGER.msg("Correctly generated 'ERROR Illegal configuration item'")
            os.chdir(initial_cwd)  # otherwise self.done does not work

        assert done, "test program should have generated an error"

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    # =====================================================================================================
    # ==================================  30 : Starting tests for TRIAGE Stage, with local ALBUM

    def program_30(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('sources/TRIAGE -> ALBUM, WEB')
        self.reset_data()
        ACTOR.rmtree("tests/results")
        ACTOR.copytree("tests/sources/TRIAGE", "tests/results/TRIAGE")
        ACTOR.copy("tests/sources/piwiPre-local.ini", "tests/results/piwiPre.ini")

        main = pwp_init(['--chdir', 'tests/results'])  # , '--debug'])
        pwp_run(main)
        self.assert_file('tests/results/ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg')     # noqa
        jpg = PwpJpg('tests/results/ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg')  # noqa
        self.assert_copyright(jpg, "(C) 2023 by Agnes BATTINI, for test")  # noqa
        self.assert_author(jpg,  "Agnes BATTINI")
        self.assert_special(jpg, "No copy allowed unless explicitly approved by Agnes BATTINI")

        self.assert_orientation(jpg, 1)

        self.assert_file('tests/results/WEB/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup-me.jpg')  # noqa
        self.assert_file('tests/results/WEB/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup-sq.jpg')  # noqa
        self.assert_file('tests/results/WEB/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup-th.jpg')  # noqa
        self.assert_file('tests/results/WEB/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup-cu_e250.jpg')  # noqa
        self.assert_file('tests/results/WEB/2023/2023-06-Juin-17-Armor-cup/index.htm')                             # noqa

        self.assert_file('tests/results/ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-45-Armor-cup.jpg')     # noqa
        self.assert_no_file('tests/results/ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-46-Armor-cup.jpg')  # noqa

        LOGGER.msg('')
        LOGGER.msg('------- end of Step 1 of pwp_test(1)')
        LOGGER.msg('')
        # 2d run of pwp_main.
        # we keep the ALBUM etc but change TRIAGE
        # we add new files
        #       Armor-cup/20230617_111349-new.jpg
        #       Armor-cup/20230617_113128-new.jpg
        #       Forêt-de-la-Corbière/IMG20230611164005-new.jpg    # noqa
        #       Forêt-de-la-Corbière/IMG20230611163210-new.jpg    # noqa
        # and keep some old ones
        #       Armor-cup/20230617_110544-copy
        #       Forêt-de-la-Corbière/20230611_170225-copy.jpg     # noqa
        #       Forêt-de-la-Corbière/IMG20230611162736-copy.jpg   # noqa

        self.reset_data()  # before any action, e.g. rmtree

        ACTOR.rmtree("tests/results/Triages")
        ACTOR.copytree("tests/sources/TRIAGE2/", "tests/results/TRIAGE")
        main = pwp_init(['--chdir', 'tests/results'])  # , '--debug'])
        pwp_run(main)
        # if the following file was present, it means that same file detection  is broken
        self.assert_no_file('tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/' +  # noqa
                       '2023-06-11-17h02-26-Forêt-de-la-Corbière.jpg')                          # noqa
        # if the following file was present,
        # it would mean that same file detection with albums is broken
        self.assert_no_file('tests/results/ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-46-Armor-cup.jpg')  # noqa
        # the 2 new files should be there
        self.assert_file('tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/' +     # noqa
                    '2023-06-11-16h32-10-Forêt-de-la-Corbière.jpg')                             # noqa
        self.assert_file('tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/' +     # noqa
                    '2023-06-11-16h40-05-Forêt-de-la-Corbière.jpg')                             # noqa

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')

        self.done(mn)

    def program_31(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('')
        LOGGER.msg('sources/TRIAGE -> ALBUM, WEB --enable-rename FALSE')
        LOGGER.msg('')

        self.reset_data()
        ACTOR.rmtree("tests/results")
        ACTOR.copytree("tests/sources/TRIAGE", "tests/results/TRIAGE")
        ACTOR.copy("tests/sources/piwiPre-local.ini", "tests/results/piwiPre.ini")

        main = pwp_init(['--chdir', 'tests/results', '--enable-rename', 'false'])  # , '--debug'])
        pwp_run(main)
        self.assert_no_file('tests/results/ALBUM/2023/2023-06-Juin-17-Armor-cup/' +                  # noqa
                            '2023-06-17-11h05-44-Armor-cup.jpg')                                     # noqa
        self.assert_file('tests/results/ALBUM/Armor-cup/20230617_110544.jpg')
        jpg = PwpJpg('tests/results/ALBUM/Armor-cup/20230617_110544.jpg')
        self.assert_orientation(jpg, 1)
        self.assert_copyright(jpg, '2023')
        self.assert_special(jpg, 'No copy allowed unless explicitly approved by')

        LOGGER.msg('')
        LOGGER.msg(f'--------------- end of  {mn}')
        self.done(mn)

    def program_32(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()   # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('management of mp4 in TRIAGE, including copyright set')
        LOGGER.msg(f"cwd = {os.getcwd()}")
        ACTOR.rmtree("tests/results")

        ACTOR.copytree(f"tests/sources/TRIAGE3/Vendée", f"tests/results/TRIAGE/Vendée")                     # noqa
        ACTOR.copy("tests/sources/piwiPre-local.ini", "tests/results/piwiPre.ini")                          # noqa
        pwp_main(['--chdir', 'tests/results'])                                                              # noqa
        photo= "tests/results/ALBUM/2023"                                                                   # noqa
        self.assert_file(f"{photo}/2023-07-Juillet-04-Vendée/2023-07-04-12h53-07-Vendée.jpg")               # noqa
        self.assert_file(f"{photo}/2023-07-Juillet-02-Vendée/2023-07-02-14h00-00-Comments.txt")             # noqa
        self.assert_file(f"{photo}/2023-01-Janvier-27-Vendée/2023-01-27-17h59-39-Vendée.mp4")             # noqa

        mp4 = PwpMp4(f"{photo}/2023-01-Janvier-27-Vendée/2023-01-27-17h59-39-Vendée.mp4")              # noqa
        self.assert_mp4_copyright(mp4, "(C) 2023 by Famille BATTINI, for test")

        web = "tests/results/WEB/2023/2023-07-Juillet-04-Vendée"                                     # noqa
        self.assert_file(f"{web}/2023-07-04-12h53-07-Vendée-th.jpg")                                 # noqa
        self.assert_file(f"{web}/2023-07-04-12h53-07-Vendée-me.jpg")                                 # noqa
        self.assert_file(f"{web}/2023-07-04-12h53-07-Vendée-cu_e250.jpg")                            # noqa

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_33(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('')
        LOGGER.msg('Verifying dates, from TRIAGE2/Forêt-de-la-Corbière')                        # noqa
        LOGGER.msg('')

        ACTOR.rmtree("tests/results")
        ACTOR.copytree("tests/sources/TRIAGE2/Forêt-de-la-Corbière",                            # noqa
                       "tests/results/TRIAGE/Forêt-de-la-Corbière")                             # noqa
        ACTOR.copytree("tests/sources/TRIAGE4/Thabor",                                          # noqa
                   "tests/results/TRIAGE/Forêt-de-la-Corbière")                                 # noqa

        ACTOR.copy("tests/sources/piwiPre-dates.ini", "tests/results/piwiPre.ini")
        pwp_main(['--chdir', 'tests/results'])

        # Forest, SM-A336B: +4h
        self.assert_file("tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +      # noqa
                         "2023-06-11-21h02-25-Forêt-de-la-Corbière.jpg")                         # noqa
        # noqa Thabor, C4100Z,C4000Z: absolute date
        self.assert_file("tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +      # noqa
                         "2023-06-11-14h24-38-Forêt-de-la-Corbière.jpg")                         # noqa
        self.assert_file("tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +      # noqa
                         "2023-06-11-14h25-06-Forêt-de-la-Corbière.jpg")                         # noqa
        # NO-DATE
        self.assert_file("tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +      # noqa
                         "Comments.txt")
        # Forest OPPOReno2 : -6h
        self.assert_file("tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +      # noqa
                         "2023-06-11-10h32-10-Forêt-de-la-Corbière.jpg")                         # noqa

        LOGGER.msg(f'--------------- end of  {mn}')
        self.done(mn)

    def program_34(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('dryrun, sources/TRIAGE -> ALBUM, WEB, album')
        ACTOR.rmtree("tests/results")
        ACTOR.copytree("tests/sources/TRIAGE", "tests/results/TRIAGE")
        ACTOR.copy("tests/sources/piwiPre-local.ini", "tests/results/piwiPre.ini")
        ACTOR.copy("tests/sources/piwiPre-corse.ini", "tests/results/TRIAGE/Corse/piwiPre.ini")         # noqa

        main = pwp_init(['--chdir', 'tests/results', '--dryrun'])  # , '--debug'])
        pwp_run(main)
        self.assert_no_dir("tests/results/ALBUM")
        self.assert_no_dir("tests/results/WEB")
        self.assert_no_dir("tests/results/BACKUP")
        self.assert_info("Would rename 'TRIAGE/Armor-cup/20230617_110544.jpg' : " +
                    "'ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg'")        # noqa

        # CAVEAT: conflict management has NOT occurred, because the files are not created in ALBUM thanks to dryrun
        #         So we do *not* know te exact thumbnail name
        #         hence the message is not accurate
        self.assert_info("Would create Thumbnail 120x120 crop=True " +
                    "for TRIAGE/Armor-cup/20230617_110544")                                    # noqa

        self.assert_info("Would create Thumbnail 120x120 crop=True " +
                         "for TRIAGE/Corse/IMG-20200517-WA0000.jpg")                           # noqa
        self.assert_info("Would create Thumbnail 120x120 crop=True " +
                         "for TRIAGE/Corse/IMG-20210818-WA0000.jpg")                           # noqa
        self.assert_info("Would create Thumbnail 120x120 crop=True " +
                         "for TRIAGE/Corse/IMG-20210819-WA0004 - Copie.jpg")                   # noqa
        self.assert_info("Would create Thumbnail 120x120 crop=True " +
                         "for TRIAGE/Corse/IMG-20210819-WA0004.jpg")                           # noqa

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_35(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('sources/TRIAGE -> ALBUM, WEB + corse.ini')      # noqa
        ACTOR.rmtree("tests/results")
        ACTOR.copytree("tests/sources/TRIAGE", "tests/results/TRIAGE")
        ACTOR.copy("tests/sources/piwiPre-local.ini", "tests/results/piwiPre.ini")
        ACTOR.copy("tests/sources/piwiPre-corse.ini", "tests/results/TRIAGE/Corse/piwiPre.ini")  # noqa

        main = pwp_init(['--chdir', 'tests/results'])  # , '--debug'])
        pwp_run(main)
        self.assert_dir("tests/results/ALBUM")
        self.assert_dir("tests/results/WEB")
        self.assert_dir("tests/results/BACKUP")
        self.assert_file("tests/results/ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg")  # noqa
        self.assert_file("tests/results/WEB/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup-sq.jpg")  # noqa

        self.assert_file("tests/results/ALBUM/Corsica/2020-05-17-12h00-00-Corsica.jpg")
        self.assert_file("tests/results/ALBUM/Corsica/2021-08-18-12h00-00-Corsica.jpg")
        self.assert_file("tests/results/ALBUM/Corsica/2021-08-19-12h04-00-Corsica.jpg")
        self.assert_file("tests/results/ALBUM/Corsica/2021-08-19-12h04-01-Corsica.jpg")

        self.assert_file("tests/results/WEB/Corsica/2021-08-19-12h04-00-Corsica-sq.jpg")
        self.assert_file("tests/results/WEB/Corsica/2021-08-19-12h04-00-Corsica-me.jpg")

        self.assert_file("tests/results/WEB/Corsica/2021-08-19-12h04-01-Corsica-th.jpg")
        self.assert_file("tests/results/WEB/Corsica/2021-08-19-12h04-01-Corsica-cu_e250.jpg")

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_36(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('sources/TRIAGE -> ALBUM, WEB + autoconfig, prepare program_37')  # noqa
        ACTOR.rmtree("tests/results")
        ACTOR.copytree("tests/sources/TRIAGE/Armor-cup", "tests/results/TRIAGE/Armor-cup")
        ACTOR.copytree("tests/sources/TRIAGE/Corse", "tests/results/TRIAGE/Corse")              # noqa

        ACTOR.copy("tests/sources/piwiPre-autoconfig.ini", "tests/results/TRIAGE/Armor-cup/piwiPre.ini")
        ACTOR.copy("tests/sources/piwiPre-autoconfig.ini", "tests/results/TRIAGE/Corse/piwiPre.ini") # noqa

        pwp_main(['--chdir', 'tests/results'])

        # we verify that all renamed images, most thumbnails and most piwiPre.ini autoconf files are there.

        self.assert_file("tests/results/ALBUM/2023/06/17/Armor-cup-001.jpg")     # noqa
        self.assert_file("tests/results/WEB/2023/06/17/Armor-cup-001-sq.jpg")    # noqa
        self.assert_file("tests/results/ALBUM/2023/06/17/piwiPre.ini")           # noqa

        self.assert_file("tests/results/ALBUM/2023/06/17/Armor-cup-002.jpg")     # noqa
        self.assert_file("tests/results/WEB/2023/06/17/Armor-cup-002-sq.jpg")    # noqa

        self.assert_no_file("tests/results/ALBUM/2023/06/17/Armor-cup-003.jpg")  # noqa

        self.assert_file("tests/results/ALBUM/2020/05/17/Corse-001.jpg")         # noqa
        self.assert_file("tests/results/WEB/2020/05/17/Corse-001-sq.jpg")        # noqa
        self.assert_file("tests/results/ALBUM/2020/05/17/piwiPre.ini")           # noqa

        self.assert_file("tests/results/ALBUM/2021/08/18/Corse-001.jpg")         # noqa
        self.assert_file("tests/results/WEB/2021/08/18/Corse-001-th.jpg")        # noqa
        self.assert_file("tests/results/ALBUM/2021/08/18/piwiPre.ini")         # noqa

        self.assert_file("tests/results/ALBUM/2021/08/19/Corse-001.jpg")         # noqa
        self.assert_file("tests/results/WEB/2021/08/19/Corse-001-me.jpg")        # noqa

        self.assert_file("tests/results/ALBUM/2021/08/19/Corse-002.jpg")         # noqa

        # verify that the.ini files have been modified with the right base:

        self.assert_file_contains("tests/results/ALBUM/2023/06/17/piwiPre.ini",
                                  "names: '{Y}/{m}/{d}/Armor-cup-{count}.{suffix}'")

        self.assert_file_contains("tests/results/ALBUM/2020/05/17/piwiPre.ini",
                                  "names: '{Y}/{m}/{d}/Corse-{count}.{suffix}'")           # noqa

        self.done(mn)

    def program_37(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        if not self.check_done('program_36'):
            self.program_36()

        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('sources/TRIAGE -> ALBUM, WEB + autoconfig')  # noqa

        # remove a thumbnail, do a verify-albums and check
        # files have not changed
        # thumbnail is there again

        ACTOR.delete("tests/results/WEB/2023/06/17/Armor-cup-001-sq.jpg")

        pwp_main(['--chdir', 'tests/results',
                  '--album', 'ALBUM',
                  '--verify-albums', '2023/06/17'])

        self.assert_file("tests/results/ALBUM/2023/06/17/Armor-cup-001.jpg")  # noqa
        self.assert_file("tests/results/WEB/2023/06/17/Armor-cup-001-sq.jpg")  # noqa
        self.assert_file("tests/results/ALBUM/2023/06/17/piwiPre.ini")  # noqa

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')

    # =====================================================================================================
    # ================================= 100: Managing the real 'network share' location

    def runner_100(self, album, web, remote_web=None):

        ACTOR.rmtree("tests/results")
        ACTOR.rmtree(album)
        ACTOR.rmtree(web)

        ACTOR.copy("tests/sources/piwiPre-mount.ini", "tests/results/piwiPre.ini")
        ACTOR.copytree("tests/sources/TRIAGE/Armor-cup", "tests/results/TRIAGE/Armor-cup")

        LOGGER.msg(f"album         = {album}")
        LOGGER.msg(f"web           = {web}")
        LOGGER.msg(f"remote web    = {remote_web}")

        remote = "true" if remote_web else 'false'

        pwp_main(['--chdir', 'tests/results', '--album', album, '--web', web, '--remote-web', str(remote_web),
                  '--enable-remote-copy', remote,
                  '--enable-piwigo-sync', 'false'])

        if album[0] != '/':
            # album is a path relative to cwd, which was changed to 'tests/results'
            album = 'tests/results/' + album
        self.assert_file(album + '/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-45-Armor-cup.jpg')  # noqa
        self.assert_file(album + '/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg')  # noqa
        self.assert_no_file(album + '/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-46-Armor-cup.jpg')  # noqa

        if web[0] != '/':
            # album is a path relative to cwd, which was changed to 'tests/results'
            web = 'tests/results/' + web
        self.assert_file(web + '/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-45-Armor-cup-me.jpg')  # noqa
        self.assert_file(web + '/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup-me.jpg')  # noqa
        self.assert_file(web + '/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-45-Armor-cup-th.jpg')  # noqa
        self.assert_file(web + '/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup-th.jpg')  # noqa
        self.assert_file(web + '/2023/2023-06-Juin-17-Armor-cup/index.htm')  # noqa

        if remote_web:
            # always an absolute path
            ACTOR.remote_file_or_dir_exists(remote_web + '/2023/2023-06-Juin-17-Armor-cup/' +           # noqa
                                            '2023-06-17-11h05-45-Armor-cup-me.jpg')
            ACTOR.remote_file_or_dir_exists(remote_web + '/2023/2023-06-Juin-17-Armor-cup/' +           # noqa
                                            '2023-06-17-11h05-44-Armor-cup-me.jpg')
            ACTOR.remote_file_or_dir_exists(remote_web + '/2023/2023-06-Juin-17-Armor-cup/' +           # noqa
                                            '2023-06-17-11h05-45-Armor-cup-th.jpg')
            ACTOR.remote_file_or_dir_exists(remote_web + '/2023/2023-06-Juin-17-Armor-cup/' +           # noqa
                                            '2023-06-17-11h05-44-Armor-cup-th.jpg')
            ACTOR.remote_file_or_dir_exists(remote_web + '/2023/2023-06-Juin-17-Armor-cup/' +           # noqa
                                            'index.htm')

        ACTOR.rmtree(album)
        ACTOR.rmtree(web)
        if remote_web:
            ACTOR.rmtree(remote_web)

    def program_100(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')

        album = 'ALBUM/tests'
        web = 'WEB/tests'

        self.runner_100(album, web)

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_101(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        # the same test as the previous, but we use the real location home_config['album']

        LOGGER.msg(f'--------------- starting {mn}')

        album = self.home_config['album'] + '/tests'
        web = self.home_config['web'] + '/tests'
        LOGGER.msg("")
        LOGGER.msg(f"tests/sources/ to '{album}' which is read from HOME/.piwiPre.ini")
        LOGGER.msg("piwigo http admin information (login, password...) SHOULD be available from there")
        LOGGER.msg("")

        self.runner_100(album, web)

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def reset_102(self):
        if os.path.isfile(self.run_file):
            ACTOR.delete(self.run_file)

        album = self.home_config['album'] + '/2023'
        if os.path.isdir(album):
            all_files = os.listdir(album)
            for f in all_files:
                m = re.match(r"2023-06-Juin-17-Armor\d+", f)
                if m:
                    ACTOR.rmtree(album + '/' + f)

        web = self.home_config['web'] + '/2023'
        if os.path.isdir(web):
            all_files = os.listdir(web)
            for f in all_files:
                m = re.match(r"Armor\d+", f)
                if m:
                    ACTOR.rmtree(web + '/' + f)

    def runner_102(self, piwigo_sync: bool):
        album = self.home_config['album']
        web = self.home_config['web']

        LOGGER.msg(f"TRIAGE3 -> {album}, piwigo sync {piwigo_sync}")
        LOGGER.msg("")
        LOGGER.msg(f"album='{album}' is read from HOME/.piwiPre.ini")
        LOGGER.msg("This SHOULD be your remote server used for test")
        LOGGER.msg("and other remote information (login, password...) SHOULD be available from there")
        LOGGER.msg("")

        if piwigo_sync and (self.home_config['piwigo-host'] is None or self.home_config['piwigo-user'] is None):
            LOGGER.msg("piwigo-host or piwigo-user not set, aborting test")
            return

        # run_file is used for debugging purposes
        # we can run runner_102 several times, it will generate files in separate directories
        # if we run program_0, e.g. for tox, then everything is reset
        if os.path.isfile(self.run_file):
            with open(self.run_file, 'r', encoding="utf-8") as f:
                all_lines = f.readlines()
            first = all_lines[0].strip()
        else:
            first = 0
        number = int(first) + 1
        with open(self.run_file, 'w', encoding="utf-8") as f:
            f.write(f"{number}\n")
            f.write("# Run number for test program 102\n")
        LOGGER.msg(f"Run number = '{number}'")
        self.reset_data()
        ACTOR.rmtree("tests/results")
        ACTOR.copytree(f"tests/sources/TRIAGE/Armor-cup", f"tests/results/TRIAGE/Armor{number}")               # noqa
        ACTOR.copy("tests/sources/piwiPre-mount.ini", "tests/results/piwiPre.ini")                           # noqa
        pwp_main(['--chdir', 'tests/results', '--enable-piwigo-sync', 'true' if piwigo_sync else 'false',     # noqa
                  '--album', album, '--web', web])

        if album[0] != '/':
            # album is a path relative to cwd, which was changed to 'tests/results'
            album = 'tests/results/' + album

        photo= f"{album}/2023/2023-06-Juin-17-Armor{number}"                                                  # noqa
        self.assert_file(f"{photo}/2023-06-17-11h05-44-Armor{number}.jpg")                                    # noqa
        self.assert_file(f"{photo}/2023-06-17-11h05-45-Armor{number}.jpg")                                    # noqa

        if web[0] != '/':
            # album is a path relative to cwd, which was changed to 'tests/results'
            web = 'tests/results/' + web

        web=f"{web}/2023/2023-06-Juin-17-Armor{number}"                                                       # noqa
        self.assert_file(f"{web}/2023-06-17-11h05-44-Armor{number}-th.jpg")                                   # noqa
        self.assert_file(f"{web}/2023-06-17-11h05-44-Armor{number}-me.jpg")                                   # noqa
        self.assert_file(f"{web}/2023-06-17-11h05-44-Armor{number}-cu_e250.jpg")                              # noqa
        self.assert_not_info("not in piwigo managed directories")
        if piwigo_sync:
            self.assert_info(f"dir '2023/2023-06-Juin-17-Armor{number}' synchronized ")                    # noqa
        return number

    def program_102(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')
        self.runner_102(False)
        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_103(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')
        self.runner_102(True)
        self.done(mn)

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.reset_done('program_102')

    # =====================================================================================================
    # ------------ 200 : Starting tests for --verify-albums

    def program_200(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('------------ Starting tests for --verify-albums')
        LOGGER.msg('')
        LOGGER.msg('sources/TRIAGE -> ALBUM, WEB')
        LOGGER.msg('')
        ACTOR.rmtree("tests/results")
        ACTOR.copytree("tests/sources/TRIAGE", "tests/results/TRIAGE")
        ACTOR.copy("tests/sources/piwiPre-local.ini", "tests/results/piwiPre.ini")
        ACTOR.copy("tests/sources/piwiPre-corse.ini", "tests/results/TRIAGE/Corse/piwiPre.ini")  # noqa

        main = pwp_init(['--chdir', 'tests/results'])  # , '--debug'])
        pwp_run(main)

        # OR: depends on the order of files processing
        self.assert_info_or(["RENAME: 'TRIAGE/Armor-cup/20230617_110544.jpg' : " +  # noqa
                            "'ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg'",  # noqa
                            "RENAME: 'TRIAGE/Armor-cup/20230617_110544.jpg' : " +  # noqa
                            "'ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-45-Armor-cup.jpg'",  # noqa
                            "New file 'TRIAGE/Armor-cup/20230617_110544.jpg' is already in album as " +
                            "'ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg'",  # noqa
                            "New file 'TRIAGE/Armor-cup/20230617_110544.jpg' is already in album as " +
                            "'ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-45-Armor-cup.jpg'"])  # noqa

        self.assert_info_or(["RENAME: 'TRIAGE/Armor-cup/20230617_110544-copy.jpg' : " +  # noqa
                             "'ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg'",  # noqa
                             "RENAME: 'TRIAGE/Armor-cup/20230617_110544-copy.jpg' : " +  # noqa
                             "'ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-45-Armor-cup.jpg'",  # noqa
                             "New file 'TRIAGE/Armor-cup/20230617_110544-copy.jpg' is already in album as " +
                             "'ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg'",  # noqa
                             "New file 'TRIAGE/Armor-cup/20230617_110544-copy.jpg' is already in album as " +
                             "'ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-45-Armor-cup.jpg'"])  # noqa

        self.assert_info_or(["RENAME: 'TRIAGE/Armor-cup/20230617_110544-bis.jpg' : " +  # noqa
                             "'ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg'",  # noqa
                             "RENAME: 'TRIAGE/Armor-cup/20230617_110544-bis.jpg' : " +  # noqa
                             "'ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-45-Armor-cup.jpg'",  # noqa
                             "New file 'TRIAGE/Armor-cup/20230617_110544-bis.jpg' is already in album as " +
                             "'ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg'",  # noqa
                             "New file 'TRIAGE/Armor-cup/20230617_110544-bis.jpg' is already in album as " +
                             "'ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-45-Armor-cup.jpg'"])  # noqa


        self.assert_info("RENAME: 'TRIAGE/Corse/IMG-20200517-WA0000.jpg' : " +  # noqa
                         "'ALBUM/Corsica/2020-05-17-12h00-00-Corsica.jpg'")  # noqa
        self.assert_info("RENAME: 'TRIAGE/Corse/IMG-20210818-WA0000.jpg' : " +  # noqa
                         "'ALBUM/Corsica/2021-08-18-12h00-00-Corsica.jpg'")  # noqa
        self.assert_info("RENAME: 'TRIAGE/Corse/IMG-20210819-WA0004 - Copie.jpg' : " +  # noqa
                         "'ALBUM/Corsica/2021-08-19-12h04-00-Corsica.jpg'")  # noqa
        self.assert_info("RENAME: 'TRIAGE/Corse/IMG-20210819-WA0004.jpg' : " +  # noqa
                         "'ALBUM/Corsica/2021-08-19-12h04-01-Corsica.jpg'")  # noqa
        self.assert_info("RENAME: 'TRIAGE/Forêt-de-la-Corbière/20230611_162803.jpg' : " +  # noqa
                         "'ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +  # noqa
                         "2023-06-11-16h28-04-Forêt-de-la-Corbière.jpg'")  # noqa
        self.assert_info("RENAME: 'TRIAGE/Forêt-de-la-Corbière/20230611_162811.jpg' : " +  # noqa
                         "'ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +  # noqa
                         "2023-06-11-16h28-11-Forêt-de-la-Corbière.jpg'")  # noqa
        self.assert_info("RENAME: 'TRIAGE/Forêt-de-la-Corbière/20230611_162816.jpg' : " +  # noqa
                         "'ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +  # noqa
                         "2023-06-11-16h28-17-Forêt-de-la-Corbière.jpg'")  # noqa
        self.assert_info("RENAME: 'TRIAGE/Forêt-de-la-Corbière/20230611_170215.jpg' : " +  # noqa
                         "'ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +  # noqa
                         "2023-06-11-17h02-16-Forêt-de-la-Corbière.jpg'")  # noqa
        self.assert_info("RENAME: 'TRIAGE/Forêt-de-la-Corbière/20230611_170225.jpg' : " +  # noqa
                         "'ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +  # noqa
                         "2023-06-11-17h02-25-Forêt-de-la-Corbière.jpg'")  # noqa
        self.assert_info("RENAME: 'TRIAGE/Forêt-de-la-Corbière/IMG20230611162736.jpg' : " +  # noqa
                         "'ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +  # noqa
                         "2023-06-11-16h27-36-Forêt-de-la-Corbière.jpg'")  # noqa
        self.assert_file('tests/results/ALBUM/Corsica/piwiPre.ini')  # noqa

        jpg = PwpJpg("tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +  # noqa
                     "2023-06-11-16h28-17-Forêt-de-la-Corbière.jpg")  # noqa
        self.assert_orientation(jpg, 1)
        self.assert_copyright(jpg, '2023')
        self.assert_special(jpg, 'No copy allowed unless explicitly approved by')
        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_201(self):
        if not self.check_done('program_200'):
            self.program_200()
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('')
        LOGGER.msg('must be executed AFTER pwp_test(200)')
        LOGGER.msg('--verify-albums, without modifications of ALBUM')

        main = pwp_init(['--chdir', 'tests/results', '--triage', 'None',
                         '--verify-albums', '2023/2023-06-Juin-11-Forêt-de-la-Corbière',   # noqa
                         '--verify-albums', '2023/2023-06-Juin-17-Armor-cup',   # noqa
                         '--verify-albums', 'Corsica'])
        pwp_run(main)
        # nothing should have changed
        self.assert_info("File 'ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière" +          # noqa
                         "/2023-06-11-16h27-36-Forêt-de-la-Corbière.jpg' has not changed")  # noqa
        self.assert_info("File 'ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière" +          # noqa
                         "/2023-06-11-16h28-04-Forêt-de-la-Corbière.jpg' has not changed")  # noqa
        self.assert_info("File 'ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière" +          # noqa
                         "/2023-06-11-16h28-11-Forêt-de-la-Corbière.jpg' has not changed")  # noqa
        self.assert_info("File 'ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière" +          # noqa
                         "/2023-06-11-16h28-17-Forêt-de-la-Corbière.jpg' has not changed")  # noqa
        self.assert_info("File 'ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière" +          # noqa
                         "/2023-06-11-17h02-16-Forêt-de-la-Corbière.jpg' has not changed")  # noqa
        self.assert_info("File 'ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière" +          # noqa
                         "/2023-06-11-17h02-25-Forêt-de-la-Corbière.jpg' has not changed")  # noqa
        self.assert_info("File 'ALBUM/2023/2023-06-Juin-17-Armor-cup" +                          # noqa
                         "/2023-06-17-11h05-44-Armor-cup.jpg' has not changed")                  # noqa
        self.assert_info("File 'ALBUM/2023/2023-06-Juin-17-Armor-cup" +                          # noqa
                         "/2023-06-17-11h05-45-Armor-cup.jpg' has not changed")                   # noqa
        self.assert_info("File 'ALBUM/Corsica/2020-05-17-12h00-00-Corsica.jpg' has not changed")  # noqa
        self.assert_info("File 'ALBUM/Corsica/2021-08-18-12h00-00-Corsica.jpg' has not changed")  # noqa
        self.assert_info("File 'ALBUM/Corsica/2021-08-19-12h04-00-Corsica.jpg' has not changed")  # noqa
        self.assert_info("File 'ALBUM/Corsica/2021-08-19-12h04-01-Corsica.jpg' has not changed")  # noqa
        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_202(self):
        if not self.check_done('program_200'):
            self.program_200()
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('')
        LOGGER.msg('--verify-albums, --dryrun, with some modifications')
        LOGGER.msg('must be executed AFTER pwp_test(200), or pwp_test(201)')
        LOGGER.msg('')
        ACTOR.rmtree("tests/results/BACKUP")

        # next picture has a bad name
        ACTOR.move('tests/results/ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg',  # noqa
                   'tests/results/ALBUM/2023/2023-06-Juin-17-Armor-cup/P1.jpg')                             # noqa
        # next picture has a bad copyright information, without '2023'
        ACTOR.delete('tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/' +                     # noqa
                     '2023-06-11-16h27-36-Forêt-de-la-Corbière.jpg')  # noqa
        ACTOR.copy('tests/sources/Modified/2023-06-11-16h27-36-Forêt-de-la-Corbière.jpg',                   # noqa
                   'tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/')                        # noqa
        # next picture has a wrong rotation
        ACTOR.delete('tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/' +                     # noqa
                     '2023-06-11-16h28-17-Forêt-de-la-Corbière.jpg')                                        # noqa
        ACTOR.copy('tests/sources/Modified/2023-06-11-16h28-17-Forêt-de-la-Corbière.jpg',                   # noqa
                   'tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/')                        # noqa

        ACTOR.delete("tests/results/WEB/Corsica/2020-05-17-12h00-00-Corsica-sq.jpg")                        # noqa

        main2 = pwp_init(['--chdir', 'tests/results', '--triage', 'None', '--dryrun',
                          '--verify-albums', '2023/2023-06-Juin-11-Forêt-de-la-Corbière',                   # noqa
                          '--verify-albums', '2023/2023-06-Juin-17-Armor-cup',                              # noqa
                          '--verify-albums', 'Corsica'])                                                    # noqa
        pwp_run(main2)
        self.assert_no_file(
            'tests/results/ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg')  # noqa
        self.assert_no_file('tests/results/WEB/Corsica/2020-05-17-12h00-00-Corsica-sq.jpg')  # noqa
        jpg = PwpJpg("tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +  # noqa
                     "2023-06-11-16h27-36-Forêt-de-la-Corbière.jpg")  # noqa
        self.assert_no_copyright(jpg, '2023')
        self.assert_no_file("tests/results/BACKUP/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +  # noqa
                            "2023-06-11-16h27-36-Forêt-de-la-Corbière.jpg")  # noqa

        jpg = PwpJpg("tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +  # noqa
                     "2023-06-11-16h28-17-Forêt-de-la-Corbière.jpg")  # noqa
        self.assert_no_orientation(jpg, 1)
        self.assert_no_file('tests/results/BACKUP/2023/2023-06-Juin-11-Forêt-de-la-Corbière/' +  # noqa
                            '2023-06-11-16h28-11-Forêt-de-la-Corbière.jpg')  # noqa                                          # noqa

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_203(self):
        if not self.check_done('program_202'):
            self.program_202()
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('')
        LOGGER.msg('--verify-albums, with some modifications, without dryrun')
        LOGGER.msg('MUST be executed AFTER pwp_test(202)')
        LOGGER.msg('')

        main2 = pwp_init(['--chdir', 'tests/results',  '--triage', 'None',
                          '--verify-albums', '2023/2023-06-Juin-11-Forêt-de-la-Corbière',                # noqa
                          '--verify-albums', '2023/2023-06-Juin-17-Armor-cup',                           # noqa
                          '--verify-albums', 'Corsica'])                                                 # noqa
        pwp_run(main2)

        # correctly renamed the file
        self.assert_file('tests/results/ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg')  # noqa
        # correctly generate lacking thumbnail
        self.assert_file('tests/results/WEB/Corsica/2020-05-17-12h00-00-Corsica-sq.jpg')  # noqa

        jpg = PwpJpg("tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +  # noqa
                     "2023-06-11-16h27-36-Forêt-de-la-Corbière.jpg")  # noqa
        self.assert_copyright(jpg, '2023')
        self.assert_file("tests/results/BACKUP/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +  # noqa
                         "2023-06-11-16h27-36-Forêt-de-la-Corbière.jpg")  # noqa

        jpg = PwpJpg("tests/results/ALBUM/2023/2023-06-Juin-11-Forêt-de-la-Corbière/" +  # noqa
                     "2023-06-11-16h28-17-Forêt-de-la-Corbière.jpg")  # noqa
        self.assert_orientation(jpg, 1)
        self.assert_file('tests/results/BACKUP/2023/2023-06-Juin-11-Forêt-de-la-Corbière/' +  # noqa
                         '2023-06-11-16h28-17-Forêt-de-la-Corbière.jpg')  # noqa                                          # noqa

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_204(self):
        if not self.check_done('program_200'):
            self.program_200()
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')

        LOGGER.msg('must be executed AFTER pwp_test(200), CAN be executed after 201/202/203')
        LOGGER.msg('--verify-album, with modification, --dryrun')

        self.create('tests/results/WEB/2023/2023-06-Juin-11-Forêt-de-la-Corbière/foo-sq.jpg')     # noqa
        self.create('tests/results/WEB/2023/2023-06-Juin-11-Forêt-de-la-Corbière/bat-th.jpg')     # noqa

        main2 = pwp_init(['--chdir', 'tests/results', '--triage', 'None', '--dryrun',
                          '--verify-albums', '2023/2023-06-Juin-11-Forêt-de-la-Corbière',                # noqa
                          '--useless-thumbnails-delete'])          # noqa
        pwp_run(main2)
        self.assert_file('tests/results/WEB/2023/2023-06-Juin-11-Forêt-de-la-Corbière/foo-sq.jpg')      # noqa
        self.assert_file('tests/results/WEB/2023/2023-06-Juin-11-Forêt-de-la-Corbière/bat-th.jpg')      # noqa

        LOGGER.msg(f'--------------- end of  {mn}')
        self.done(mn)

    def program_205(self):
        if not self.check_done('program_204'):
            self.program_204()
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')

        LOGGER.msg('must be executed AFTER pwp_test(204)')
        LOGGER.msg('--verify-album, with modification, WITHOUT dryrun')

        main2 = pwp_init(['--chdir', 'tests/results', '--triage', 'None',
                          '--verify-albums', '2023/2023-06-Juin-11-Forêt-de-la-Corbière',                 # noqa
                          '--useless-thumbnails-delete'])
        pwp_run(main2)
        self.assert_no_file('tests/results/WEB/2023/2023-06-Juin-11-Forêt-de-la-Corbière/foo-sq.jpg')     # noqa
        self.assert_no_file('tests/results/WEB/2023/2023-06-Juin-11-Forêt-de-la-Corbière/bat-th.jpg')     # noqa
        self.reset_done('program_204')
        self.reset_done('program_203')
        self.reset_done('program_202')
        self.reset_done('program_201')
        self.reset_done('program_200')
        LOGGER.msg(f'--------------- end of  {mn}')
        self.done(mn)

    def program_206(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')

        LOGGER.msg('')
        LOGGER.msg('--building the test case with tests/sources/TRIAGE/Armor-cup ')
        LOGGER.msg('')
        ACTOR.rmtree("tests/results")
        ACTOR.copytree("tests/sources/TRIAGE/Armor-cup", "tests/results/TRIAGE/Armor-cup")
        ACTOR.copy("tests/sources/piwiPre-local.ini", "tests/results/piwiPre.ini")

        # first, we generate the test pattern in ALBUM
        pwp_main(['--chdir', 'tests/results', '--enable-rename', 'true'])

        self.assert_file('tests/results/ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-44-Armor-cup.jpg')  # noqa
        self.assert_file('tests/results/WEB/2023/2023-06-Juin-17-Armor-cup/' +                                    # noqa
                         '2023-06-17-11h05-44-Armor-cup-sq.jpg')                                                  # noqa

        self.assert_file('tests/results/ALBUM/2023/2023-06-Juin-17-Armor-cup/2023-06-17-11h05-45-Armor-cup.jpg')  # noqa
        self.assert_no_file('tests/results/WEB/2023/2023-06-Juin-17-Armor-cup/' +                                 # noqa
                            '2023-06-17-11h05-46-Armor-cup.jpg')                                                  # noqa

        LOGGER.msg(f'--------------- end of  {mn}')
        self.done(mn)

    def program_207(self):
        if not self.check_done('program_206'):
            self.program_206()
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('')
        LOGGER.msg('--verify-album --enable-rename , with a different names ')
        LOGGER.msg('')

        ACTOR.copy('tests/sources/piwiPre-names.ini', 'tests/results/piwiPre.ini')
        self.reset_done('program_206')

        pwp_main(['--chdir', 'tests/results', '--triage', 'None',
                  '--verify-albums', '2023/2023-06-Juin-17-Armor-cup',                      # noqa
                  '--useless-thumbnails-delete'])

        self.assert_no_file('tests/results/ALBUM/2023/2023-06-Juin-17-Armor-cup/' +         # noqa
                            '2023-06-17-11h05-44-Armor-cup.jpg')                            # noqa
        self.assert_no_file('tests/results/WEB/2023/2023-06-Juin-17-Armor-cup/' +           # noqa
                         '2023-06-17-11h05-44-Armor-cup-sq.jpg')                            # noqa

        self.assert_file('tests/results/ALBUM/2023/06/17/Armor-cup-002.jpg')                # noqa

        LOGGER.msg(f'--------------- end of  {mn}')
        self.done(mn)

    def program_210(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('--verify-album --dryrun in real life')

        album = self.home_config['album']
        target = self.get_first_album()
        if target is None:
            LOGGER.msg(f"album '{album}' is empty, no test to do")
        else:
            LOGGER.msg(f"verifying album '{target}' --dryrun")

            ACTOR.rmtree("tests/results")
            ACTOR.mkdirs("tests/results")
            ACTOR.copy("tests/sources/piwiPre-mount.ini", "tests/results/piwiPre.ini")
            pwp_main(['--chdir', 'tests/results',
                      '--verify-albums', target, "--dryrun"])
        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    # ====================================================================
    # --verify-thumbnails

    def program_300(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree
        if not self.check_done('program_30'):
            self.program_30()

        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('verify-thumbnails')

        ACTOR.delete("tests/results/WEB/2023/2023-06-Juin-17-Armor-cup/" +                               # noqa
                     "2023-06-17-11h05-44-Armor-cup-cu_e250.jpg")                                        # noqa
        ACTOR.delete("tests/results/WEB/2021/2021-08-Août-19-Corse/2021-08-19-12h04-00-Corse-sq.jpg")    # noqa
        self.create("tests/results/WEB/2021/2021-08-Août-19-Corse/2021-08-19-12h04-00-Corse-foo.jpg")    # noqa
        self.create("tests/results/WEB/2020/2020-05-Mai-17-Corse/2021-08-19-12h04-00-Corse-sq.jpg")      # noqa
        self.create("tests/results/WEB/2020/2020-05-Mai-17-Corse/foo.txt")                               # noqa

        pwp_main(['--chdir', 'tests/results',
                  '--verify-thumbnails', "2020",
                  '--verify-thumbnails', "2021",
                  '--verify-thumbnails', "2023",
                  '--useless-thumbnails-delete'])  # noqa

        self.assert_file("tests/results/WEB/2023/2023-06-Juin-17-Armor-cup/" +              # noqa
                         "2023-06-17-11h05-44-Armor-cup-cu_e250.jpg")                       # noqa
        self.assert_file("tests/results/WEB/2021/2021-08-Août-19-Corse/" +                  # noqa
                         "2021-08-19-12h04-00-Corse-sq.jpg")                                # noqa
        self.assert_no_file("tests/results/WEB/2021/2021-08-Août-19-Corse/" +               # noqa
                         "2021-08-19-12h04-00-Corse-foo.jpg")                               # noqa
        self.assert_no_file("tests/results/WEB/2020/2020-05-Mai-17-Corse/" +                # noqa
                         "2021-08-19-12h04-00-Corse-sq.jpg")                                # noqa
        self.assert_no_file("tests/results/WEB/2020/2020-05-Mai-17-Corse/foo.txt")          # noqa

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    # ====================================================================
    # ssh/sftp

    def program_400(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('ssh')

        ACTOR.rmtree("tests/results")
        ACTOR.copy("tests/sources/piwiPre-mount.ini", "tests/results/piwiPre.ini")
        self.create("tests/results/dummy.txt")

        pwp_main(['--chdir', 'tests/results', '--test-ssh', '--test-sftp', '--enable-remote-copy', 'true'])  # noqa

        self.assert_not_info("SSH error")
        self.assert_info("sftp test OK")

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    def program_401(self):
        mn = inspect.getframeinfo(inspect.currentframe()).function
        self.reset_data()  # before any action, e.g. rmtree

        LOGGER.msg(f'--------------- starting {mn}')
        LOGGER.msg('remote web, with TRIAGE')

        ACTOR.rmtree("tests/results")

        album = self.home_config['album'] + '/tests'
        web = 'WEB/tests'
        remote_web = self.home_config['remote-web']   # + '/tests'
        self.runner_100(album, web, remote_web=remote_web)

        LOGGER.msg(f'--------------- end of  {mn}')
        LOGGER.msg('')
        self.done(mn)

    # ====================================================================

    def run_number(self, number, running_all):
        initial_cwd = os.getcwd()
        name = f"program_{number}"

        try:
            if name in self.programs:
                self.programs[name]()
            else:
                if not running_all:
                    LOGGER.msg('')
                    LOGGER.msg(f'---- No program {name}')
                    LOGGER.msg('')
        # except PwpError:  # traps all errors and subclasses
        #    LOGGER.msg("continuing tests after PwpError")
        except SystemExit:
            LOGGER.msg("continuing tests after SystemExit")
        os.chdir(initial_cwd)

    def program_0(self):
        LOGGER.msg('----------------------------- starting program_0')
        LOGGER.msg(f"cwd  = '{os.getcwd()}' ")
        LOGGER.msg(f"HOME = '{os.path.expanduser('~')}' ")

        self.reset_102()
        for i in range(1, self.last+1):  # in [400]:  # for i in [3, 400]:  # for i in [3, 15, 17, 400]
            self.run_number(i, running_all=True)

        LOGGER.msg('----------------------------- end program_0')
        LOGGER.msg('')
        LOGGER.msg('End of ALL TESTS')
        LOGGER.msg('')
        end = datetime.datetime.now()

        LOGGER.msg(f"--- asserts      = {self.asserts}")
        LOGGER.msg(f"--- end          = {end} ---")
        LOGGER.msg(f"--- files        = {self.pictures_nb}")
        LOGGER.msg(f"--- duration     = {end - self.start_time}")
        if self.pictures_nb:
            LOGGER.msg(f"--- duration/pic = {(end - self.start_time) / self.pictures_nb}")
        LOGGER.msg("------------------------------------")


def test_all_programs():  # because it is called test_XXX, tox runs it
    all_tests = PwpTester()
    all_tests.program_0()


def run_pwp(arguments):
    LOGGER.msg('--------------- starting run_pwp()')
    parser = argparse.ArgumentParser(description='testing piwiPre ')
    parser.add_argument('--number', '-n', help='test number', action='store')
    args = parser.parse_args() if arguments is None else parser.parse_args(arguments)

    number = int(args.number) if args.number else 0

    all_tests = PwpTester()
    all_tests.run_number(number, running_all=False)


if __name__ == '__main__':
    run_pwp(sys.argv[1:])
